import enum
import os
import shlex
from functools import partial

from libqtile.command.base import expose_command
from libqtile.log_utils import logger
from libqtile.widget import base

BACKLIGHT_DIR = "/sys/class/backlight"


@enum.unique
class ChangeDirection(enum.Enum):
    UP = 0
    DOWN = 1


def find_default_backlight() -> str:
    try:
        entries = os.listdir(BACKLIGHT_DIR)
        if entries:
            return entries[0]
    except FileNotFoundError:
        pass
    return "QTILE_BACKLIGHT_NOT_FOUND"


class Backlight(base.InLoopPollText):
    """A simple widget to show the current brightness of a monitor.

    If the change_command parameter is set to None, the widget will attempt to
    use the interface at /sys/class to change brightness. This depends on
    having the correct udev rules, so be sure Qtile's udev rules are installed
    correctly.

    You can also bind keyboard shortcuts to the backlight widget with:

    .. code-block:: python

        from libqtile.widget import backlight
        Key(
            [],
            "XF86MonBrightnessUp",
            lazy.widget['backlight'].change_backlight(backlight.ChangeDirection.UP)
        )
        Key(
            [],
            "XF86MonBrightnessDown",
            lazy.widget['backlight'].change_backlight(backlight.ChangeDirection.DOWN)
        )
    """

    filenames: dict = {}

    defaults = [
        ("backlight_name", find_default_backlight(), "ACPI name of a backlight device"),
        (
            "brightness_file",
            "brightness",
            "Name of file with the current brightness in /sys/class/backlight/backlight_name",
        ),
        (
            "max_brightness_file",
            "max_brightness",
            "Name of file with the maximum brightness in /sys/class/backlight/backlight_name",
        ),
        ("update_interval", 0.2, "The delay in seconds between updates"),
        ("step", 10, "Percent of backlight every scroll changed"),
        ("format", "{percent:2.0%}", "Display format"),
        ("change_command", "xbacklight -set {0}", "Execute command to change value"),
        ("min_brightness", 0, "Minimum brightness percentage"),
    ]

    def __init__(self, **config):
        base.InLoopPollText.__init__(self, **config)
        self.add_defaults(Backlight.defaults)
        self._future = None

        self.brightness_file = os.path.join(
            BACKLIGHT_DIR,
            self.backlight_name,
            self.brightness_file,
        )
        self.max_brightness_file = os.path.join(
            BACKLIGHT_DIR,
            self.backlight_name,
            self.max_brightness_file,
        )

        self.add_callbacks(
            {
                "Button4": partial(self.change_backlight, ChangeDirection.UP),
                "Button5": partial(self.change_backlight, ChangeDirection.DOWN),
            }
        )

    def finalize(self):
        if self._future and not self._future.done():
            self._future.cancel()
        base.InLoopPollText.finalize(self)

    def _load_file(self, path):
        try:
            with open(path) as f:
                return float(f.read().strip())
        except FileNotFoundError:
            logger.debug("Failed to get %s", path)
            raise RuntimeError(f"Unable to read status for {os.path.basename(path)}")

    def _get_info(self):
        brightness = self._load_file(self.brightness_file)
        max_value = self._load_file(self.max_brightness_file)
        return brightness / max_value

    def poll(self):
        try:
            percent = self._get_info()
        except RuntimeError as e:
            return f"Error: {e}"

        return self.format.format(percent=percent)

    def _change_backlight(self, value):
        if self.change_command is None:
            value = self._load_file(self.max_brightness_file) * value / 100
            try:
                with open(self.brightness_file, "w") as f:
                    f.write(str(round(value)))
            except PermissionError:
                logger.warning(
                    "Cannot set brightness: no write permission for %s", self.brightness_file
                )
        else:
            self.call_process(shlex.split(self.change_command.format(value)))

    @expose_command()
    def change_backlight(self, direction, step=None):
        if not step:
            step = self.step
        if self._future and not self._future.done():
            return
        new = now = self._get_info() * 100
        if direction is ChangeDirection.DOWN:
            new = max(now - step, self.min_brightness)
        elif direction is ChangeDirection.UP:
            new = min(now + step, 100)
        if new != now:
            self._future = self.qtile.run_in_executor(self._change_backlight, new)



import libqtile.layout
from libqtile import hook
from libqtile.lazy import lazy
from libqtile.widget import TextBox


class ScreenSplit(TextBox):
    """
    A simple widget to show the name of the current split and layout for the
    ``ScreenSplit`` layout.
    """

    defaults = [("format", "{split_name} ({layout})", "Format string.")]

    def __init__(self, **config):
        TextBox.__init__(self, "", **config)
        self.add_defaults(ScreenSplit.defaults)
        self.add_callbacks(
            {
                "Button4": lazy.layout.previous_split().when(layout="screensplit"),
                "Button5": lazy.layout.next_split().when(layout="screensplit"),
            }
        )
        self._drawn = False

    def _configure(self, qtile, bar):
        TextBox._configure(self, qtile, bar)
        self._set_hooks()

    def _set_hooks(self):
        hook.subscribe.layout_change(self._update)

    def _update(self, layout, group):
        if not self.configured:
            return

        if isinstance(layout, libqtile.layout.ScreenSplit) and group is self.bar.screen.group:
            split_name = layout.active_split.name
            layout_name = layout.active_layout.name
            self.set_text(split_name, layout_name)
        else:
            self.clear()

    def set_text(self, split_name, layout):
        self.update(self.format.format(split_name=split_name, layout=layout))

    def clear(self):
        self.update("")

    def finalize(self):
        hook.unsubscribe.layout_change(self._update)
        TextBox.finalize(self)

    def draw(self):
        # Force the widget to check layout the first time it's drawn
        if not self._drawn:
            self._update(self.bar.screen.group.layout, self.bar.screen.group)
            self._drawn = True
            return

        TextBox.draw(self)



import asyncio
import contextlib
from enum import Enum

from dbus_fast.aio import MessageBus
from dbus_fast.constants import BusType
from dbus_fast.errors import DBusError, InterfaceNotFoundError

from libqtile.command.base import expose_command
from libqtile.log_utils import logger
from libqtile.utils import create_task
from libqtile.widget import base

BLUEZ_SERVICE = "org.bluez"
BLUEZ_DEVICE = "org.bluez.Device1"
BLUEZ_ADAPTER = "org.bluez.Adapter1"
BLUEZ_BATTERY = "org.bluez.Battery1"
OBJECT_MANAGER_INTERFACE = "org.freedesktop.DBus.ObjectManager"
PROPERTIES_INTERFACE = "org.freedesktop.DBus.Properties"


def _catch_dbus_error(msg):
    """Decorator to catch DBusErrors and log a message."""

    def _wrapper(func):
        async def f(self):
            try:
                await func(self)
            except DBusError:
                logger.warning(msg, self.name)

        return f

    return _wrapper


class DeviceState(Enum):
    CONNECTED = 0
    PAIRED = 1
    UNPAIRED = 2


class _BluetoothBase:
    """Base class with some common requirements for devices and adapters."""

    def __init__(self, path, interface, properties_interface, widget):
        self.path = path
        self.interface = interface
        self.widget = widget
        self.properties = properties_interface
        self.properties.on_properties_changed(self.properties_changed)
        self._name = ""

    @property
    def name(self):
        return self._name

    def __repr__(self):
        """Neater repr to help debugging."""
        return f"<{self.__class__.__name__}: {self.name} ({self.path})>"

    def __del__(self):
        """Remove signal listener when garbage collected."""
        with contextlib.suppress(RuntimeError):
            self.properties.off_properties_changed(self.properties_changed)

    def properties_changed(self, _interface_name, _changed_properties, _invalidated_properties):
        """Handler for properties_changed signal."""
        create_task(self.update_props())

    async def update_props(self):
        raise NotImplementedError


class BluetoothDevice(_BluetoothBase):
    """
    Helper class to represent an org.bluez.Device1 object.

    Exposes basic properties/methods and also listens to signals to
    update properties as needed.
    """

    def __init__(self, path, interface, properties_interface, widget):
        _BluetoothBase.__init__(self, path, interface, properties_interface, widget)
        self._connected = False
        self._paired = False
        self._status = DeviceState.UNPAIRED
        self._adapter = ""
        self.has_name = False
        self.battery_device = None
        self._battery = 0

    @_catch_dbus_error("Unable to connect to device: %s.")
    async def connect(self):
        await self.interface.call_connect()

    @_catch_dbus_error("Unable to disconnect device: %s.")
    async def disconnect(self):
        await self.interface.call_disconnect()

    @_catch_dbus_error("Unable to pair device: %s.")
    async def pair_and_connect(self):
        await self.interface.call_pair()
        await self.connect()

    async def action(self):
        """Helper method to call appropriate method based on device status."""
        if self.connected:
            await self.disconnect()
        elif self.paired and not self.connected:
            await self.connect()
        elif not self.paired:
            await self.pair_and_connect()

    @property
    def connected(self):
        return self._connected

    @property
    def paired(self):
        return self._paired

    @property
    def status(self):
        return self._status

    @property
    def battery(self):
        if self.battery_device:
            return self._battery
        return ""

    @property
    def adapter(self):
        a = self.widget.adapters.get(self._adapter)
        if a:
            return a.name
        else:
            return "Unknown"

    def add_battery(self):
        """Triggers adding battery interface."""

        def refresh(_):
            if self.battery_device:
                create_task(self.update_props())

        task = create_task(self.get_battery())
        task.add_done_callback(refresh)

    def remove_battery(self):
        self.battery_device = None
        self._battery = 0

    async def get_battery(self):
        proxy = await self.widget.get_proxy(self.path)
        with contextlib.suppress(InterfaceNotFoundError):
            self.battery_device = proxy.get_interface(BLUEZ_BATTERY)

    async def update_props(self, setup=False):
        """Refresh all the properties for the device."""
        # Some devices don't report a name so we fall back to the device address
        try:
            self._name = await self.interface.get_name()
            self.has_name = True
        except (AttributeError, DBusError):
            self._name = await self.interface.get_address()
            self.has_name = False

        self._connected, self._paired, self._adapter = await asyncio.gather(
            self.interface.get_connected(),
            self.interface.get_paired(),
            self.interface.get_adapter(),
        )

        # If we're setting up, let's see if a battery device is available
        # This may happen if the device is already connected when the widget starts
        if setup:
            await self.get_battery()

        if self.battery_device:
            self._battery = await self.battery_device.get_percentage()

        if self._connected:
            self._status = DeviceState.CONNECTED
        elif self._paired and not self._connected:
            self._status = DeviceState.PAIRED
        else:
            self._status = DeviceState.UNPAIRED

        if not setup:
            self.widget.refresh()

    async def check(self):
        """Checks if device belongs to requested adapter."""
        await self.update_props(setup=True)

        if not self.widget.adapter_paths:
            return True, self

        for path in self.widget.adapter_paths:
            if path == self._adapter:
                return True, self

        return False, self


class BluetoothAdapter(_BluetoothBase):
    """
    Helper class for Bluetooth adapters.

    Exposes basic properties/methods and also listens to signals to
    update properties as needed.
    """

    def __init__(self, path, interface, properties_interface, widget):
        _BluetoothBase.__init__(self, path, interface, properties_interface, widget)
        self._discovering = False
        self._powered = False
        create_task(self.update_props(setup=True))

    @_catch_dbus_error("Unable to start discovery on adapter: %s.")
    async def start_discovery(self):
        await self.interface.call_start_discovery()

    @_catch_dbus_error("Unable to stop discovery on adapter: %s.")
    async def stop_discovery(self):
        await self.interface.call_stop_discovery()

    @_catch_dbus_error("Unable to set power state for adapter: %s.")
    async def power(self):
        await self.interface.set_powered(not self._powered)

    @property
    def discovering(self):
        return self._discovering

    @property
    def powered(self):
        return self._powered

    async def discover(self):
        if self.discovering:
            await self.stop_discovery()
        else:
            await self.start_discovery()

    async def update_props(self, setup=False):
        self._discovering = await self.interface.get_discovering()
        self._powered = await self.interface.get_powered()
        self._name = await self.interface.get_name()

        if not setup:
            self.widget.refresh()


class Bluetooth(base._TextBox):
    """
    Bluetooth widget that provides following functionality:
    - View multiple adapters/devices (adapters can be filtered)
    - Set power and discovery status for adapters
    - Connect/disconnect/pair devices

    The widget works by providing a menu in the bar. Different items are accessed
    by scrolling up and down on the widget.

    Clicking on an adapter will open a submenu allowing you to set power and discovery status.

    Clicking on a device will perform an action based on the status of that device:
    - Connected devices will be disconnected
    - Disconnected devices will be connected
    - Unpaired devices (which appear if discovery is on) will be paired and connected

    Symbols are used to show the status of adapters and devices.

    Battery level for bluetooth devices can also be shown if available. This functionality is not available
    by default on all distros. If it doesn't work, you can try adding ``Experimental = true`` to
    ``/etc/bluetooth/main.conf``.
    """

    defaults = [
        ("hide_unnamed_devices", False, "Devices with no name will be hidden from scan results"),
        ("symbol_connected", "*", "Symbol to indicate device is connected"),
        ("symbol_paired", "-", "Symbol to indicate device is paired but unconnected"),
        ("symbol_unknown", "?", "Symbol to indicate device is unpaired"),
        ("symbol_powered", ("*", "-"), "Symbols when adapter is powered and unpowered."),
        (
            "symbol_discovery",
            ("D", ""),
            "Symbols when adapter is discovering and not discovering",
        ),
        (
            "device_format",
            "Device: {name}{battery_level} [{symbol}]",
            "Text to display when showing bluetooth device. "
            "The ``{adapter`` field is also available if you're using multiple adapters.",
        ),
        (
            "device_battery_format",
            " ({battery}%)",
            "Text to be shown if device reports battery level",
        ),
        (
            "adapter_format",
            "Adapter: {name} [{powered}{discovery}]",
            "Text to display when showing adapter device.",
        ),
        (
            "adapter_paths",
            [],
            "List of DBus object path for bluetooth adapter (e.g. '/org/bluez/hci0'). "
            "Empty list will show all adapters.",
        ),
        (
            "default_text",
            "BT {connected_devices}",
            "Text to show when not scrolling through menu. "
            "Available fields: 'connected_devices' list of connected devices, "
            "'num_connected_devices' number of connected devices, "
            "'adapters' list of bluetooth adapters, 'num_adapters' number of bluetooth adapters.",
        ),
        (
            "default_show_battery",
            False,
            "Include battery level of 'connected_devices' in 'default_text'. Uses 'device_battery_format'.",
        ),
        ("separator", ", ", "Separator for lists in 'default_text'."),
        (
            "default_timeout",
            None,
            "Time before reverting to default_text. If 'None', text will stay on selected item.",
        ),
        (
            "device",
            None,
            "Device path, can be found with d-feet or similar dbus explorer. "
            "When set, the widget will default to showing this device status.",
        ),
    ]

    def __init__(self, **config):
        base._TextBox.__init__(self, **config)
        self.add_defaults(Bluetooth.defaults)
        self.connected = False
        self.bus = None
        self.devices = {}
        self.adapters = {}
        self._lines = []
        self._line_index = 0
        self._adapter_index = 0
        self._setting_up = True
        self.show_adapter = False
        self.device_found = False
        self.add_callbacks(
            {"Button1": self.click, "Button4": self.scroll_up, "Button5": self.scroll_down}
        )
        self.timer = None
        self.object_manager = None

    def _configure(self, qtile, bar):
        base._TextBox._configure(self, qtile, bar)
        self.symbols = (self.symbol_connected, self.symbol_paired, self.symbol_unknown)

    async def _config_async(self):
        await self._connect()

    async def _connect(self):
        """Connect to bus and set up key listeners."""
        self.bus = await MessageBus(bus_type=BusType.SYSTEM).connect()

        # Get the object manager
        proxy = await self.get_proxy("/")
        self.object_manager = proxy.get_interface(OBJECT_MANAGER_INTERFACE)

        # Subscribe to signals for new and removed interfaces
        self.object_manager.on_interfaces_added(self._interface_added)
        self.object_manager.on_interfaces_removed(self._interface_removed)

        await self._get_managed_objects()

        self.refresh()

    async def get_proxy(self, path):
        """Provides proxy object after introspecting the given path."""
        device_introspection = await self.bus.introspect(BLUEZ_SERVICE, path)
        proxy = self.bus.get_proxy_object(BLUEZ_SERVICE, path, device_introspection)
        return proxy

    async def _get_managed_objects(self):
        """
        Retrieve list of managed objects.

        These are devices that have previously been paired but may or may
        not currently be connected.

        Additionally, if the device is scanning, available objects will also
        appear here, albeit temporarily.
        """
        self._setting_up = True

        objects = await self.object_manager.call_get_managed_objects()

        for path, interfaces in objects.items():
            self._interface_added(path, interfaces)

        self._setting_up = False

    def _interface_added(self, path, interfaces):
        """Handles the object based on the interface type."""
        task = None

        # Create device or adapter
        for device_type in (BLUEZ_DEVICE, BLUEZ_ADAPTER):
            if device_type in interfaces:
                task = create_task(self._add_object(path, device_type))
                break

        if task is not None and not self._setting_up:
            task.add_done_callback(lambda *args: self.refresh())

        # Battery interface is added after the device is connected
        # so no task will have been created at this point.
        if not task and BLUEZ_BATTERY in interfaces:
            if device := self.devices.get(path):
                # Get device to load battery interface
                device.add_battery()

    def _interface_removed(self, path, interfaces):
        # Object has been removed so remove from our list of available devices
        updated = False

        if BLUEZ_DEVICE in interfaces:
            with contextlib.suppress(KeyError):
                del self.devices[path]
                updated = True

        elif BLUEZ_ADAPTER in interfaces:
            with contextlib.suppress(KeyError):
                del self.adapters[path]
                updated = True

        elif BLUEZ_BATTERY in interfaces:
            device = self.devices.get(path)
            if device:
                device.remove_battery()

        if updated and not self._setting_up:
            self.refresh()

    async def _add_object(self, path, device_type):
        proxy = await self.get_proxy(path)

        # Check if object is a valid bluetooth device and ignore it if not
        try:
            interface = proxy.get_interface(device_type)
        except InterfaceNotFoundError:
            return
        # Get the properties interface so we can listed to signals
        properties = proxy.get_interface(PROPERTIES_INTERFACE)

        # Create an object to represent this device and add to our list
        if device_type == BLUEZ_DEVICE:
            # The device will be added to self.devices as part of the __init__
            # process after checking whether we need to filter out the device
            # if it's connected to an unwatched adapter
            # This is preferable to substring matching on 'path' which could
            # result in false positives (in very rare situations)
            device = BluetoothDevice(path, interface, properties, self)
            task = create_task(device.check())
            task.add_done_callback(self._add_device)
        elif device_type == BLUEZ_ADAPTER:
            if not self.adapter_paths or path in self.adapter_paths:
                adapter = BluetoothAdapter(path, interface, properties, self)
                self.adapters[path] = adapter

    def _add_device(self, task):
        success, device = task.result()
        if success:
            self.devices[device.path] = device
            self.refresh()

    def refresh(self):
        if self._setting_up:
            return

        # Store lines in a variable as we'll need to access them elsewhere
        # Each entry is a tuple of (display text, callable)
        self._lines = []

        # If we've clicked on an adapter then we're just showing the adapter submenu
        if self.show_adapter:
            self._lines.extend(self._get_adapter_menu(self._shown_adapter))

        # Otherwise we're in default behavior
        else:
            # Line 1 is the text to be formatted according to "default_text"
            connected = [d for d in self.devices.values() if d.connected]
            adapters = [a.name for a in self.adapters.values()]
            if self.default_show_battery:
                connected_devices = [
                    "{name}{battery}".format(
                        name=d.name,
                        battery=self.device_battery_format.format(battery=d.battery)
                        if d.battery
                        else "",
                    )
                    for d in connected
                ]
            else:
                connected_devices = [d.name for d in connected]
            self._lines.append(
                (
                    self.default_text.format(
                        connected_devices=self.separator.join(connected_devices),
                        num_connected_devices=len(connected_devices),
                        adapters=self.separator.join(adapters),
                        num_adapters=len(adapters),
                    ),
                    lambda: None,
                )
            )

            # Next is the adapters...
            def show(adapter):
                """Function to trigger the adapter submenu."""
                self.show_adapter = True
                self._shown_adapter = adapter
                # Store the current menu position
                self._adapter_index = self._line_index
                # Change menu position to the first item in the submenu
                self._line_index = 0
                self.refresh()

            for adapter in self.adapters.values():
                self._lines.append((adapter, lambda a=adapter: show(a)))

            # Finally, loop over all the devices
            for device in self.devices.values():
                self._lines.append((device, lambda d=device: create_task(d.action())))

        if self._lines:
            # If user has set default device, check if it should be shown
            # This will only force the display to that widget the first time the device is found
            # i.e. once user has scrolled to a different device, it will no longer return to the
            # set device.
            if not self.device_found and self.device is not None:
                for i, (obj, _) in enumerate(self._lines):
                    if isinstance(obj, BluetoothDevice):
                        if self.device in obj.path:
                            self._line_index = i
                            self.device_found = True
                            break

            self.show_line()
        else:
            self.update("")

    def _get_adapter_menu(self, adapter):
        """Builds a submenu for the selected adapter."""
        state = "off" if adapter.powered else "on"
        discovery = "off" if adapter.discovering else "on"

        def exit():
            self.show_adapter = False
            # Restore menu position
            self._line_index = self._adapter_index
            self.refresh()

        return [
            (f"Turn power {state}", lambda a=adapter: create_task(a.power())),
            (f"Turn discovery {discovery}", lambda a=adapter: create_task(a.discover())),
            ("Exit", lambda: exit()),
        ]

    def show_line(self):
        """Formats the text of the current menu item."""
        if not self._lines:
            return

        obj = None

        # If devices disappear we may have an invalid line index
        while obj is None:
            try:
                obj, action = self._lines[self._line_index]
            except IndexError:
                self._line_index -= 1

        self.update(self.format_object(obj))

    def format_object(self, obj):
        """Takes the given object and returns a formatted string representing the object."""
        if isinstance(obj, BluetoothDevice):
            # status.value is 0 for connected, 1 for paired (and unconnected), 2 for unpaired
            symbol = self.symbols[obj.status.value]
            if obj.battery:
                battery_level = self.device_battery_format.format(battery=obj.battery)
            else:
                battery_level = ""
            return self.device_format.format(
                symbol=symbol, name=obj.name, adapter=obj.adapter, battery_level=battery_level
            )

        elif isinstance(obj, BluetoothAdapter):
            powered = 0 if obj.powered else 1
            discovery = 0 if obj.discovering else 1
            return self.adapter_format.format(
                powered=self.symbol_powered[powered],
                discovery=self.symbol_discovery[discovery],
                name=obj.name,
            )

        elif isinstance(obj, str):
            return obj

        # Shouldn't happen but let's be safe!
        return ""

    @expose_command
    def scroll_up(self):
        """Scroll up to next item."""
        self._scroll(1)

    @expose_command
    def scroll_down(self):
        """Scroll down to next item."""
        self._scroll(-1)

    @expose_command
    def click(self):
        """Perform default action on visible item."""
        with contextlib.suppress(IndexError):
            _, action = self._lines[self._line_index]
            action()

    def _scroll(self, step):
        if self.timer is not None:
            self.timer.cancel()

        if self._lines:
            self._line_index = (self._line_index + step) % len(self._lines)
            self.show_line()

        if self.default_timeout is not None:
            self.timer = self.timeout_add(self.default_timeout, self.hide)

    def hide(self):
        """Revert widget contents to default."""
        self._line_index = 0
        self.show_adapter = False
        self.refresh()

    def finalize(self):
        # if we failed to connect, there is nothing to finalize.
        if self.bus is None:
            return

        # Remove dbus signal handlers before finalising.
        # Clearing dicts will call the __del__ method on the stored objects
        # which has been defined to remove signal handlers
        self.devices.clear()
        self.adapters.clear()

        # Remove object manager's handlers
        if self.object_manager is not None:
            self.object_manager.off_interfaces_added(self._interface_added)
            self.object_manager.off_interfaces_removed(self._interface_removed)

        # Disconnect the bus connection
        self.bus.disconnect()
        self.bus = None

        base._TextBox.finalize(self)



from os import path

from libqtile import bar, pangocffi, utils
from libqtile.command.base import expose_command
from libqtile.log_utils import logger
from libqtile.notify import ClosedReason, notifier
from libqtile.widget import base


class Notify(base._TextBox):
    """
    A notify widget

    This widget can handle actions provided by notification clients. However, only the
    default action is supported, so if a client provides multiple actions then only the
    default (first) action can be invoked. Some programs will provide their own
    notification windows if the notification server does not support actions, so if you
    want your notifications to handle more than one action then specify ``False`` for
    the ``action`` option to disable all action handling. Unfortunately we cannot
    specify the capability for exactly one action.
    """

    defaults = [
        ("foreground_urgent", "ff0000", "Foreground urgent priority colour"),
        ("foreground_low", "dddddd", "Foreground low priority  colour"),
        ("default_timeout_low", 5, "Default timeout (seconds) for low urgency notifications."),
        ("default_timeout", 10, "Default timeout (seconds) for normal notifications"),
        ("default_timeout_urgent", None, "Default timeout (seconds) for urgent notifications"),
        ("audiofile", None, "Audiofile played during notifications"),
        ("action", True, "Enable handling of default action upon right click"),
        (
            "parse_text",
            None,
            "Function to parse and modify notifications. "
            "e.g. function in config that removes line returns:"
            "def my_func(text)"
            "   return text.replace('\n', '')"
            "then set option parse_text=my_func",
        ),
        ("background_urgent", "440000", "Background urgent priority colour"),
        ("background_low", "444444", "Background low priority colour"),
    ]
    capabilities = {"body", "actions"}

    def __init__(self, width=bar.CALCULATED, **config):
        base._TextBox.__init__(self, "", width, **config)
        self.add_defaults(Notify.defaults)
        self.current_id = 0

        default_callbacks = {
            "Button1": self.clear,
            "Button4": self.prev,
            "Button5": self.next,
        }
        if self.action:
            default_callbacks["Button3"] = self._invoke
        else:
            self.capabilities = Notify.capabilities.difference({"actions"})
        self.add_callbacks(default_callbacks)

        self.background_normal = self.background

    def _configure(self, qtile, bar):
        base._TextBox._configure(self, qtile, bar)
        self.layout = self.drawer.textlayout(
            self.text, self.foreground, self.font, self.fontsize, self.fontshadow, markup=True
        )
        if notifier is None:
            logger.warning("You must install dbus-fast to use the Notify widget.")

        # Create a tuple of our default timeouts. Urgency is an integer of 0-2
        # (see https://specifications.freedesktop.org/notification-spec/notification-spec-latest.html#urgency-levels)
        # so they will work as the index of the tuple.
        self._timeouts = (
            self.default_timeout_low,
            self.default_timeout,
            self.default_timeout_urgent,
        )

    async def _config_async(self):
        if notifier is None:
            return

        await notifier.register(self.update, self.capabilities, on_close=self.on_close)

    def set_notif_text(self, notif):
        self.text = pangocffi.markup_escape_text(notif.summary)
        urgency = getattr(notif.hints.get("urgency"), "value", 1)
        if urgency != 1:
            self.text = f'<span color="{utils.hex(self.foreground_urgent if urgency == 2 else self.foreground_low)}">{self.text}</span>'
            self.background = self.background_urgent if urgency == 2 else self.background_low
        else:
            self.background = self.background_normal

        if notif.body:
            self.text = f'<span weight="bold">{self.text}</span> - {pangocffi.markup_escape_text(notif.body)}'
        if callable(self.parse_text):
            try:
                self.text = self.parse_text(self.text)
            except:  # noqa: E722
                logger.exception("parse_text function failed:")
        if self.audiofile and path.exists(self.audiofile):
            self.qtile.spawn(f"aplay -q '{self.audiofile}'")

    def update(self, notif):
        self.qtile.call_soon_threadsafe(self.real_update, notif)

    def real_update(self, notif):
        self.set_notif_text(notif)
        self.current_id = notif.id - 1
        if notif.timeout and notif.timeout > 0:
            self.timeout_add(
                notif.timeout / 1000, self.clear, method_args=(ClosedReason.expired,)
            )
        else:
            urgency = getattr(notif.hints.get("urgency"), "value", 1)
            try:
                timeout = self._timeouts[urgency]
            except IndexError:
                logger.warning(
                    "Notification had an unexpected urgency value. Treating as normal priority."
                )
                timeout = self._timeouts[1]

            if timeout:
                self.timeout_add(timeout, self.clear, method_args=(ClosedReason.expired,))
        self.bar.draw()
        return True

    @expose_command()
    def display(self):
        if notifier is None:
            return

        self.set_notif_text(notifier.notifications[self.current_id])
        self.bar.draw()

    @expose_command()
    def clear(self, reason=ClosedReason.dismissed):
        """Clear the notification"""
        if notifier is None:
            return

        notifier._service.NotificationClosed(notifier.notifications[self.current_id].id, reason)
        self.text = ""
        self.background = self.background_normal
        self.current_id = len(notifier.notifications) - 1
        self.bar.draw()

    def on_close(self, nid):
        if self.current_id < len(notifier.notifications):
            notif = notifier.notifications[self.current_id]
            if notif.id == nid:
                self.clear(ClosedReason.method)

    @expose_command()
    def prev(self):
        """Show previous notification."""
        if self.current_id > 0:
            self.current_id -= 1
        self.display()

    @expose_command()
    def next(self):
        if notifier is None:
            return

        """Show next notification."""
        if self.current_id < len(notifier.notifications) - 1:
            self.current_id += 1
            self.display()

    def _invoke(self):
        if self.current_id < len(notifier.notifications):
            notif = notifier.notifications[self.current_id]
            if notif.actions:
                notifier._service.ActionInvoked(notif.id, notif.actions[0])
            self.clear()

    @expose_command()
    def toggle(self):
        """Toggle showing/clearing the notification"""
        if self.text == "":
            self.display()
        else:
            self.clear()

    @expose_command()
    def invoke(self):
        """Invoke the notification's default action"""
        if self.action:
            self._invoke()

    def finalize(self):
        if notifier is not None:
            notifier.unregister(self.update, on_close=self.on_close)
        base._TextBox.finalize(self)



"""
The objects defining the nodes in the command graph and the navigation of the
abstract command graph
"""

from __future__ import annotations

import abc
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    SelectorType = tuple[str, str | int | None]


class CommandGraphNode(metaclass=abc.ABCMeta):
    """A container node in the command graph structure

    A command graph node which can contain other elements that it can link to.
    May also have commands that can be executed on itself.
    """

    @property
    @abc.abstractmethod
    def selector(self) -> str | int | None:
        """The selector for the current node"""

    @property
    @abc.abstractmethod
    def selectors(self) -> list[SelectorType]:
        """The selectors resolving the location of the node in the command graph"""

    @property
    @abc.abstractmethod
    def parent(self) -> CommandGraphNode | None:
        """The parent of the current node"""

    @property
    @abc.abstractmethod
    def children(self) -> list[str]:
        """The child objects that are contained within this object"""

    def navigate(self, name: str, selector: str | int | None) -> CommandGraphNode:
        """Navigate from the current node to the specified child"""
        if name in self.children:
            return _COMMAND_GRAPH_MAP[name](selector, self)
        raise KeyError(f"Given node is not an object: {name}")

    def call(self, name: str, lifted: bool = False) -> CommandGraphCall:
        """Execute the given call on the selected object"""
        return CommandGraphCall(name, self, lifted=lifted)


class CommandGraphCall:
    """A call performed on a particular object in the command graph"""

    def __init__(self, name: str, parent: CommandGraphNode, lifted: bool = False) -> None:
        """A command to be executed on the selected object

        A terminal node in the command graph, specifying an actual command to
        execute on the selected graph element.

        Parameters
        ----------
        name:
            The name of the command to execute
        parent:
            The command graph node on which to execute the given command.
        """
        self._name = name
        self._parent = parent
        self.lifted = lifted

    @property
    def name(self) -> str:
        """The name of the call to make"""
        return self._name

    @property
    def selectors(self) -> list[SelectorType]:
        """The selectors resolving the location of the node in the command graph"""
        return self.parent.selectors

    @property
    def parent(self) -> CommandGraphNode:
        """The parent of the current node"""
        return self._parent


class CommandGraphRoot(CommandGraphNode):
    """The root node of the command graph

    Contains all of the elements connected to the root of the command graph.
    """

    @property
    def selector(self) -> None:
        """The selector for the current node"""
        return None

    @property
    def selectors(self) -> list[SelectorType]:
        """The selectors resolving the location of the node in the command graph"""
        return []

    @property
    def parent(self) -> None:
        """The parent of the current node"""
        return None

    @property
    def children(self) -> list[str]:
        """All of the child elements in the root of the command graph"""
        return ["bar", "group", "layout", "screen", "widget", "window", "core"]


class CommandGraphObject(CommandGraphNode, metaclass=abc.ABCMeta):
    """An object in the command graph that contains a collection of objects"""

    def __init__(self, selector: str | int | None, parent: CommandGraphNode) -> None:
        """A container object in the command graph

        Parameters
        ----------
        selector: str | None
            The name of the selected element within the command graph.  If not
            given, corresponds to the default selection of this type of object.
        parent: CommandGraphNode
            The container object that this object is the child of.
        """
        self._selector = selector
        self._parent = parent

    @property
    def selector(self) -> str | int | None:
        """The selector for the current node"""
        return self._selector

    @property
    def selectors(self) -> list[SelectorType]:
        """The selectors resolving the location of the node in the command graph"""
        selectors = self.parent.selectors + [(self.object_type, self.selector)]
        return selectors

    @property
    def parent(self) -> CommandGraphNode:
        """The parent of the current node"""
        return self._parent

    @property
    @abc.abstractmethod
    def object_type(self) -> str:
        """The type of the current container object"""


class _BarGraphNode(CommandGraphObject):
    object_type = "bar"
    children = ["screen", "widget"]


class _GroupGraphNode(CommandGraphObject):
    object_type = "group"
    children = ["layout", "window", "screen"]


class _LayoutGraphNode(CommandGraphObject):
    object_type = "layout"
    children = ["group", "window", "screen"]


class _ScreenGraphNode(CommandGraphObject):
    object_type = "screen"
    children = ["layout", "window", "bar", "widget", "group"]


class _WidgetGraphNode(CommandGraphObject):
    object_type = "widget"
    children = ["bar", "screen"]


class _WindowGraphNode(CommandGraphObject):
    object_type = "window"
    children = ["group", "screen", "layout"]


class _CoreGraphNode(CommandGraphObject):
    object_type = "core"
    children: list[str] = []


_COMMAND_GRAPH_MAP: dict[str, type[CommandGraphObject]] = {
    "bar": _BarGraphNode,
    "group": _GroupGraphNode,
    "layout": _LayoutGraphNode,
    "widget": _WidgetGraphNode,
    "window": _WindowGraphNode,
    "screen": _ScreenGraphNode,
    "core": _CoreGraphNode,
}


GraphType = CommandGraphNode | CommandGraphCall



import json
from typing import Any

import aiohttp

from libqtile.log_utils import logger
from libqtile.widget import base

try:
    import xmltodict

    def xmlparse(body):
        return xmltodict.parse(body)

except ImportError:
    # TODO: we could implement a similar parser by hand, but i'm lazy, so let's
    # punt for now
    def xmlparse(body):
        raise Exception("no xmltodict library")


class GenPollUrl(base.BackgroundPoll):
    """A generic text widget that polls an url and parses it using parse function


    Widget requirements: aiohttp_.

    .. _aiohttp: https://pypi.org/project/aiohttp/
    """

    defaults: list[tuple[str, Any, str]] = [
        ("url", None, "Url"),
        ("data", None, "Post Data"),
        ("parse", None, "Parse Function"),
        ("json", True, "Is Json?"),
        ("user_agent", "Qtile", "Set the user agent"),
        ("headers", {}, "Extra Headers"),
        ("xml", False, "Is XML?"),
    ]

    def __init__(self, **config):
        base.BackgroundPoll.__init__(self, "", **config)
        self.add_defaults(GenPollUrl.defaults)

        self.headers["User-agent"] = self.user_agent
        if self.json:
            self.headers["Content-Type"] = "application/json"

        if self.data and not isinstance(self.data, str):
            self.data = json.dumps(self.data).encode()

    async def apoll(self):
        if not self.parse or not self.url:
            return "Invalid config"

        headers = self.headers.copy()
        data = self.data

        try:
            async with aiohttp.ClientSession() as session:
                async with session.request(
                    method="POST" if data else "GET", url=self.url, data=data, headers=headers
                ) as response:
                    if self.json:
                        body = await response.json()
                    elif self.xml:
                        text_body = await response.text()
                        body = xmlparse(text_body)
                    else:
                        body = await response.text()

            text = self.parse(body)
        except Exception:
            logger.exception("got exception polling widget")
            text = "Can't parse"

        return text



from __future__ import annotations

from typing import TYPE_CHECKING

from libqtile.command.base import expose_command
from libqtile.log_utils import logger
from libqtile.pangocffi import markup_escape_text
from libqtile.widget import Systray, base

if TYPE_CHECKING:
    from typing import Any


class WidgetBox(base._TextBox):
    """A widget to declutter your bar.

    WidgetBox is a widget that hides widgets by default but shows them when
    the box is opened.

    Widgets that are hidden will still update etc. as if they were on the main
    bar.

    Button clicks are passed to widgets when they are visible so callbacks will
    work.

    Widgets in the box also remain accessible via command interfaces.

    Widgets can only be added to the box via the configuration file. The widget
    is configured by adding widgets to the "widgets" parameter as follows::

        widget.WidgetBox(widgets=[
            widget.TextBox(text="This widget is in the box"),
            widget.Memory()
            ]
        ),
    """

    orientations = base.ORIENTATION_HORIZONTAL
    defaults: list[tuple[str, Any, str]] = [
        (
            "close_button_location",
            "left",
            "Location of close button when box open ('left' or 'right')",
        ),
        ("text_closed", "[<]", "Text when box is closed"),
        ("text_open", "[>]", "Text when box is open"),
        ("widgets", list(), "A list of widgets to include in the box"),
        ("start_opened", False, "Spawn the box opened"),
    ]

    def __init__(self, _widgets: list[base._Widget] | None = None, **config):
        base._TextBox.__init__(self, **config)
        self.add_defaults(WidgetBox.defaults)
        self.box_is_open = False
        self.add_callbacks({"Button1": self.toggle})

        if _widgets:
            logger.warning(
                "The use of a positional argument in WidgetBox is deprecated. "
                "Please update your config to use widgets=[...]."
            )
            self.widgets = _widgets

        self.close_button_location: str
        if self.close_button_location not in ["left", "right"]:
            val = self.close_button_location
            logger.warning("Invalid value for 'close_button_location': %s", val)
            self.close_button_location = "left"

    def _configure(self, qtile, bar):
        base._TextBox._configure(self, qtile, bar)

        self.text = markup_escape_text(self.text_open if self.box_is_open else self.text_closed)

        if self.configured:
            return

        for idx, w in enumerate(self.widgets):
            if w.configured:
                w = w.create_mirror()
                self.widgets[idx] = w
            self.qtile.register_widget(w)
            w._configure(self.qtile, self.bar)
            w.offsety = self.bar.border_width[0]

            # In case the widget is mirrored, we need to draw it once so the
            # mirror can copy the surface but draw it off screen
            w.offsetx = self.bar.length
            self.qtile.call_soon(w.draw)

            # Setting the configured flag for widgets was moved to Bar._configure so we need to
            # set it here.
            w.configured = True

        # Disable drawing of the widget's contents
        for w in self.widgets:
            w.drawer.disable()

        # We're being cautious: `box_is_open` should never be True here...
        if self.start_opened and not self.box_is_open:
            self.qtile.call_soon(self.toggle)

    def set_box_label(self):
        self.text = markup_escape_text(self.text_open if self.box_is_open else self.text_closed)

    def toggle_widgets(self):
        for widget in self.widgets:
            try:
                self.bar.widgets.remove(widget)
                # Override drawer.drawer with a no-op
                widget.drawer.disable()

                # Systray widget needs some additional steps to hide as the icons
                # are separate _Window instances.
                # Systray unhides icons when it draws so we only need to hide them.
                if isinstance(widget, Systray):
                    for icon in widget.tray_icons:
                        icon.hide()

            except ValueError:
                continue

        index = self.bar.widgets.index(self)

        if self.close_button_location == "left":
            index += 1

        if self.box_is_open:
            # Need to reverse list as widgets get added in front of eachother.
            for widget in self.widgets[::-1]:
                # enable drawing again
                widget.drawer.enable()
                self.bar.widgets.insert(index, widget)

    @expose_command()
    def toggle(self):
        """Toggle box state"""
        self.box_is_open = not self.box_is_open
        self.toggle_widgets()
        self.set_box_label()
        self.bar.draw()

    @expose_command()
    def open(self):
        """Open the widgetbox."""
        if not self.box_is_open:
            self.toggle()

    @expose_command()
    def close(self):
        """Close the widgetbox."""
        if self.box_is_open:
            self.toggle()



import os
from pathlib import Path

import cairocffi

try:
    from xdg.DesktopEntry import DesktopEntry
    from xdg.IconTheme import getIconPath

    has_xdg = True
except ImportError:
    has_xdg = False

import libqtile.bar
from libqtile import hook, pangocffi
from libqtile.images import Img
from libqtile.log_utils import logger
from libqtile.widget import base

data_home = os.environ.get("XDG_DATA_HOME", "~/.local/share")

DESKTOP_LOCATIONS = [
    Path(data_home) / "applications",
    Path("/usr/local/share/applications"),
    Path("/usr/share/applications"),
]


class TaskList(base._Widget, base.PaddingMixin, base.MarginMixin):
    """Displays the icon and name of each window in the current group

    Contrary to WindowTabs this is an interactive widget.  The window that
    currently has focus is highlighted.

    Optional requirements: `pyxdg <https://pypi.org/project/pyxdg/>`__ is needed
    to use theme icons and to display icons on Wayland.
    """

    orientations = base.ORIENTATION_BOTH
    defaults = [
        ("font", "sans", "Default font"),
        ("fontsize", None, "Font size. Calculated if None."),
        ("foreground", "ffffff", "Foreground colour"),
        ("fontshadow", None, "font shadow color, default is None(no shadow)"),
        ("borderwidth", 2, "Current group border width"),
        ("border", "215578", "Border colour"),
        ("rounded", True, "To round or not to round borders"),
        (
            "highlight_method",
            "border",
            "Method of highlighting (one of 'border' or 'block') Uses `*_border` color settings",
        ),
        ("urgent_border", "FF0000", "Urgent border color"),
        (
            "urgent_alert_method",
            "border",
            "Method for alerting you of WM urgent hints (one of 'border' or 'text')",
        ),
        (
            "unfocused_border",
            None,
            "Border color for unfocused windows. "
            "Affects only hightlight_method 'border' and 'block'. "
            "Defaults to None, which means no special color.",
        ),
        (
            "max_title_width",
            None,
            "Max size in pixels of task title.(if set to None, as much as available.)",
        ),
        (
            "title_width_method",
            None,
            "Method to compute the width of task title. (None, 'uniform'.)"
            "Defaults to None, the normal behaviour.",
        ),
        (
            "parse_text",
            None,
            "Function to parse and modify window names. "
            "e.g. function in config that removes excess "
            "strings from window name: "
            "def my_func(text)"
            '    for string in [" - Chromium", " - Firefox"]:'
            '        text = text.replace(string, "")'
            "   return text"
            "then set option parse_text=my_func",
        ),
        (
            "spacing",
            None,
            "Spacing between tasks. If set to None, defaults to margin_x in "
            "horizontal bars and margin_y in vertical bars.",
        ),
        (
            "txt_minimized",
            "_ ",
            'Text representation of the minimized window state. e.g., "_ " or "\U0001f5d5 "',
        ),
        (
            "txt_maximized",
            "[] ",
            'Text representation of the maximized window state. e.g., "[] " or "\U0001f5d6 "',
        ),
        (
            "txt_floating",
            "V ",
            'Text representation of the floating window state. e.g., "V " or "\U0001f5d7 "',
        ),
        (
            "markup_normal",
            None,
            "Text markup of the normal window state. Supports pangomarkup with markup=True."
            'e.g., "{}" or "<span underline="low">{}</span>"',
        ),
        (
            "markup_minimized",
            None,
            "Text markup of the minimized window state. Supports pangomarkup with markup=True."
            'e.g., "{}" or "<span underline="low">{}</span>"',
        ),
        (
            "markup_maximized",
            None,
            "Text markup of the maximized window state. Supports pangomarkup with markup=True."
            'e.g., "{}" or "<span underline="low">{}</span>"',
        ),
        (
            "markup_floating",
            None,
            "Text markup of the floating window state. Supports pangomarkup with markup=True."
            'e.g., "{}" or "<span underline="low">{}</span>"',
        ),
        (
            "markup_focused",
            None,
            "Text markup of the focused window state. Supports pangomarkup with markup=True."
            'e.g., "{}" or "<span underline="low">{}</span>"',
        ),
        (
            "markup_focused_floating",
            None,
            "Text markup of the focused and floating window state. Supports pangomarkup with markup=True."
            'e.g., "{}" or "<span underline="low">{}</span>"',
        ),
        (
            "icon_size",
            None,
            "Icon size. (Calculated if set to None. Icons are hidden if set to 0.)",
        ),
        (
            "theme_mode",
            None,
            "When to use theme icons. `None` = never, `preferred` = use if available, "
            "`fallback` = use if app does not provide icon directly. "
            "`preferred` and `fallback` have identical behaviour on Wayland.",
        ),
        (
            "theme_path",
            None,
            "Path to icon theme to be used by pyxdg for icons. ``None`` will use default icon theme.",
        ),
        (
            "window_name_location",
            False,
            "Whether to show the location of the window in the title.",
        ),
        (
            "window_name_location_offset",
            0,
            "The offset given to the window location",
        ),
        (
            "stretch",
            True,
            "Widget fills available space in bar. Set to `False` to limit widget width to size of its contents.",
        ),
    ]

    def __init__(self, **config):
        base._Widget.__init__(self, libqtile.bar.STRETCH, **config)
        self.add_defaults(TaskList.defaults)
        self._icons_cache = {}
        self._box_end_positions = []
        self.markup = False
        self.clicked = None

        self.add_callbacks({"Button1": self.select_window})

    def box_width(self, text):
        """
        Calculate box width for given text.
        If max_title_width is given, the returned width is limited to it.
        """
        width, height = self.drawer.max_layout_size(
            [text], self.font, self.fontsize, markup=self.markup
        )
        width = width + 2 * (self.padding_side + self.borderwidth)
        return width

    def get_taskname(self, window):
        """
        Get display name for given window.
        Depending on its state minimized, maximized and floating
        appropriate characters are prepended.
        """
        state = ""
        markup_str = self.markup_normal

        if window is None:
            pass
        elif window.minimized:
            state = self.txt_minimized
            markup_str = self.markup_minimized
        elif window.maximized:
            state = self.txt_maximized
            markup_str = self.markup_maximized
        elif window is window.group.current_window:
            if window.floating:
                state = self.txt_floating
                markup_str = self.markup_focused_floating or self.markup_floating
            else:
                markup_str = self.markup_focused
        elif window.floating:
            state = self.txt_floating
            markup_str = self.markup_floating

        window_location = (
            f"[{window.group.windows.index(window) + self.window_name_location_offset}] "
            if self.window_name_location
            else ""
        )
        window_name = window_location + window.name if window and window.name else "?"

        if callable(self.parse_text):
            try:
                window_name = self.parse_text(window_name)
            except:  # noqa: E722
                logger.exception("parse_text function failed:")

        # Emulate default widget behavior if markup_str is None
        if self.markup and markup_str is None:
            markup_str = f"{state}{{}}"

        if markup_str is not None:
            window_name = pangocffi.markup_escape_text(window_name)
            return markup_str.format(window_name)

        return f"{state}{window_name}"

    @property
    def windows(self):
        windows = []
        if self.qtile.core.name == "x11":
            for w in self.bar.screen.group.windows:
                wm_states = list(w.window.get_property("_NET_WM_STATE", "ATOM", unpack=int))
                skip_taskbar = w.qtile.core.conn.atoms["_NET_WM_STATE_SKIP_TASKBAR"]
                if w.window.get_wm_type() in ("normal", None) and skip_taskbar not in wm_states:
                    windows.append(w)
        else:
            for w in self.bar.screen.group.windows:
                if w.get_wm_type() in ("normal", None) and not w.skip_taskbar:
                    windows.append(w)

        return windows

    @property
    def max_width(self):
        width = self.bar.length - sum(
            w.length
            for w in self.bar.widgets
            if w is not self and w.length_type != libqtile.bar.STRETCH
        )
        return width

    def calc_box_widths(self):
        """
        Calculate box width for each window in current group.
        If the available space is less than overall size of boxes,
        the boxes are shrunk by percentage if greater than average.
        """
        windows = self.windows
        window_count = len(windows)

        # if no windows present for current group just return empty list
        if not window_count:
            return []

        # Determine available and max average width for task name boxes.
        width_total = self.max_width - 2 * self.margin_side - (window_count - 1) * self.spacing
        width_avg = width_total / window_count

        names = [self.get_taskname(w) for w in windows]

        if self.icon_size == 0:
            icons = len(windows) * [None]
        else:
            icons = [self.get_window_icon(w) for w in windows]

        # Obey title_width_method if specified
        if self.title_width_method == "uniform":
            width_uniform = width_total // window_count
            width_boxes = [width_uniform for w in range(window_count)]
        else:
            # Default behaviour: calculated width for each task according to
            # icon and task name consisting
            # of state abbreviation and window name
            width_boxes = [
                (
                    self.box_width(names[idx])
                    + ((self.icon_size + self.padding_side) if icons[idx] else 0)
                )
                for idx in range(window_count)
            ]

        # Obey max_title_width if specified
        if self.max_title_width:
            width_boxes = [min(w, self.max_title_width) for w in width_boxes]

        width_sum = sum(width_boxes)

        # calculated box width are to wide for available widget space:
        if width_sum > width_total:
            # sum the width of tasks shorter than calculated average
            # and calculate a ratio to shrink boxes greater than width_avg
            width_shorter_sum = sum([w for w in width_boxes if w < width_avg])

            ratio = (width_total - width_shorter_sum) / (width_sum - width_shorter_sum)
            # determine new box widths by shrinking boxes greater than avg
            width_boxes = [(w if w < width_avg else w * ratio) for w in width_boxes]

        return zip(windows, icons, names, width_boxes)

    def calculate_length(self):
        width = 0
        box_widths = [box[3] for box in self.calc_box_widths()]
        if box_widths:
            width += self.spacing * len(box_widths) - 1
            width += sum(w for w in box_widths)

        return width

    def _configure(self, qtile, bar):
        base._Widget._configure(self, qtile, bar)

        if self.spacing is None:
            self.spacing = self.margin_side

        if not self.stretch:
            self.length_type = libqtile.bar.CALCULATED

        if not has_xdg and self.theme_mode is not None:
            logger.warning("You must install pyxdg to use theme icons.")
            self.theme_mode = None

        if self.theme_mode and self.theme_mode not in ["preferred", "fallback"]:
            logger.warning(
                "Unexpected theme_mode (%s). Theme icons will be disabled.", self.theme_mode
            )
            self.theme_mode = None

        if qtile.core.name == "wayland" and self.theme_mode is None and self.icon_size != 0:
            # Disable icons
            self.icon_size = 0

        if self.icon_size is None:
            self.icon_size = self.bar.size - 2 * (self.borderwidth + self.margin_top)

        if self.fontsize is None:
            calc = (
                self.bar.size - self.margin_top * 2 - self.borderwidth * 2 - self.padding_top * 2
            )
            self.fontsize = max(calc, 1)

        # Enforce markup and new string format behaviour when
        # at least one markup_* option is used.
        # Mixing non markup and markup may cause problems.
        self.markup = bool(
            self.markup_normal
            or self.markup_minimized
            or self.markup_maximized
            or self.markup_floating
            or self.markup_focused
            or self.markup_focused_floating
        )
        self.layout = self.drawer.textlayout(
            "",
            "ffffff",
            self.font,
            self.fontsize,
            self.fontshadow,
            wrap=False,
            markup=self.markup,
        )
        self.setup_hooks()

    def update(self, window=None):
        if not window or window in self.windows:
            self.bar.draw()

    def remove_icon_cache(self, window):
        wid = window.wid
        if wid in self._icons_cache:
            self._icons_cache.pop(wid)

    def invalidate_cache(self, window):
        self.remove_icon_cache(window)
        self.update(window)

    def setup_hooks(self):
        hook.subscribe.client_name_updated(self.update)
        hook.subscribe.focus_change(self.update)
        hook.subscribe.float_change(self.update)
        hook.subscribe.client_urgent_hint_changed(self.update)

        hook.subscribe.net_wm_icon_change(self.invalidate_cache)
        hook.subscribe.client_killed(self.remove_icon_cache)

    def drawtext(self, text, textcolor, width):
        self.layout.text = text
        self.layout.colour = textcolor
        if width is not None:
            self.layout.width = width

    def drawbox(
        self,
        offset,
        text,
        bordercolor,
        textcolor,
        width=None,
        rounded=False,
        block=False,
        icon=None,
    ):
        self.drawtext(text, textcolor, width)

        icon_padding = (self.icon_size + self.padding_side) if icon else 0
        pad_x = [self.padding_side + icon_padding, self.padding_side]

        if bordercolor is None:
            # border colour is set to None when we don't want to draw a border at all
            # Rather than dealing with alpha blending issues, we just set border width
            # to 0.
            border_width = 0
            framecolor = self.background or self.bar.background
        else:
            border_width = self.borderwidth
            framecolor = bordercolor

        framed = self.layout.framed(border_width, framecolor, pad_x, self.padding_top)
        if block and bordercolor is not None:
            framed.draw_fill(offset, self.margin_top, rounded)
        else:
            framed.draw(offset, self.margin_top, rounded)

        if icon:
            self.draw_icon(icon, offset)

    def get_clicked(self, x, y):
        box_start = self.margin_side
        if self.bar.horizontal:
            pos = x
        elif self.bar.screen.left is self.bar:
            pos = self.length - y
        else:
            pos = y
        for box_end, win in zip(self._box_end_positions, self.windows):
            if box_start <= pos <= box_end:
                return win
            else:
                box_start = box_end + self.spacing
        # not found any , return None
        return None

    def button_press(self, x, y, button):
        self.clicked = self.get_clicked(x, y)
        base._Widget.button_press(self, x, y, button)

    def select_window(self):
        if self.clicked:
            current_win = self.bar.screen.group.current_window
            window = self.clicked
            if window is not current_win:
                window.group.focus(window, False)
                if window.floating:
                    window.bring_to_front()
            else:
                window.toggle_minimize()

    def _get_class_icon(self, window):
        if not getattr(window, "icons", False):
            return None

        icons = sorted(
            iter(window.icons.items()),
            key=lambda x: abs(self.icon_size - int(x[0].split("x")[0])),
        )
        icon = icons[0]
        width, height = map(int, icon[0].split("x"))

        img = cairocffi.ImageSurface.create_for_data(
            icon[1], cairocffi.FORMAT_ARGB32, width, height
        )

        return img

    def _read_desktop_file(self, desktopfile):
        icon = DesktopEntry(desktopfile).getIcon()
        if icon and Path(icon).expanduser().exists():
            return icon

    def _find_desktop_entry(self, classes):
        for p in DESKTOP_LOCATIONS:
            for cl in classes:
                for app in set([cl, cl.lower()]):
                    f = (p / f"{app}.desktop").expanduser()
                    if f.exists():
                        icon = self._read_desktop_file(f)
                        if icon:
                            return icon

    def _get_desktop_icon(self, window):
        classes = window.get_wm_class()

        if not classes:
            return None

        icon = self._find_desktop_entry(classes)

        if icon:
            img = Img.from_path(icon)
            return img.surface

    def _get_theme_icon(self, window):
        classes = window.get_wm_class()

        if not classes:
            return None

        icon = None

        for cl in classes:
            for app in set([cl, cl.lower()]):
                icon = getIconPath(app, theme=self.theme_path)
                if icon is not None:
                    break
            else:
                continue
            break

        if not icon:
            return None

        img = Img.from_path(icon)

        return img.surface

    def get_window_icon(self, window):
        if not getattr(window, "icons", False) and self.theme_mode is None:
            return None

        cache = self._icons_cache.get(window.wid)
        if cache:
            return cache

        surface = None
        img = None

        if self.qtile.core.name == "x11":
            img = self._get_class_icon(window)

        if img is None and has_xdg and self.theme_mode != "preferred":
            img = self._get_desktop_icon(window)

        if self.theme_mode == "preferred" or (self.theme_mode == "fallback" and img is None):
            xdg_img = self._get_theme_icon(window)
            if xdg_img:
                img = xdg_img

        if img is not None:
            surface = cairocffi.SurfacePattern(img)
            height = img.get_height()
            width = img.get_width()
            scaler = cairocffi.Matrix()
            if height != self.icon_size:
                sp = height / self.icon_size
                height = self.icon_size
                width /= sp
                scaler.scale(sp, sp)
            surface.set_matrix(scaler)

        self._icons_cache[window.wid] = surface
        return surface

    def draw_icon(self, surface, offset):
        if not surface:
            return

        x = offset + self.borderwidth + self.padding_side
        size = self.height if self.bar.horizontal else self.width
        y = (size - self.icon_size) // 2

        self.drawer.ctx.save()
        self.drawer.ctx.translate(x, y)
        self.drawer.ctx.set_source(surface)
        self.drawer.ctx.paint()
        self.drawer.ctx.restore()

    def draw(self):
        self.drawer.clear(self.background or self.bar.background)
        offset = self.margin_side
        self.drawer.ctx.save()
        self.rotate_drawer()

        self._box_end_positions = []
        for w, icon, task, bw in self.calc_box_widths():
            self._box_end_positions.append(offset + bw)

            if w.urgent:
                border = self.urgent_border
                text_color = border
            elif w is w.group.current_window:
                border = self.border
                text_color = border
            else:
                border = self.unfocused_border or None
                text_color = self.foreground

            if self.highlight_method == "text":
                border = None
            else:
                text_color = self.foreground

            textwidth = (
                bw - 2 * self.padding_side - ((self.icon_size + self.padding_side) if icon else 0)
            )
            self.drawbox(
                offset,
                task,
                border,
                text_color,
                rounded=self.rounded,
                block=(self.highlight_method == "block"),
                width=textwidth,
                icon=icon,
            )
            offset += bw + self.spacing

        self.drawer.ctx.restore()
        self.draw_at_default_position()



from __future__ import annotations

import math
from typing import TYPE_CHECKING

from libqtile.command.base import expose_command
from libqtile.layout.base import _SimpleLayoutBase

if TYPE_CHECKING:
    from typing import Any, Self

    from libqtile.backend.base import Window
    from libqtile.config import ScreenRect
    from libqtile.group import _Group


class Matrix(_SimpleLayoutBase):
    """
    This layout divides the screen into a matrix of equally sized cells and
    places one window in each cell. The number of columns is configurable and
    can also be changed interactively.
    """

    columns: int

    defaults = [
        ("border_focus", "#0000ff", "Border colour(s) for the focused window."),
        ("border_normal", "#000000", "Border colour(s) for un-focused windows."),
        ("border_width", 1, "Border width."),
        ("columns", 2, "Number of columns"),
        ("margin", 0, "Margin of the layout (int or list of ints [N E S W])"),
    ]

    def __init__(self, **config):
        _SimpleLayoutBase.__init__(self, **config)
        self.add_defaults(Matrix.defaults)

    @property
    def rows(self):
        """Calc current number of rows, basd on number of clients and columns"""
        return int(math.ceil(len(self.clients) / self.columns))

    @property
    def row(self):
        """Calc row index of current client"""
        return self.clients.current_index // self.columns

    @property
    def column(self):
        """Calc column index of current client"""
        return self.clients.current_index % self.columns

    @expose_command()
    def info(self) -> dict[str, Any]:
        d = _SimpleLayoutBase.info(self)
        d["rows"] = [[win.name for win in self.get_row(i)] for i in range(self.rows)]
        d["current_window"] = self.column, self.row
        return d

    def clone(self, group: _Group) -> Self:
        c = _SimpleLayoutBase.clone(self, group)
        c.columns = self.columns
        return c

    def get_row(self, row):
        """Get all clients in given row"""
        assert row < self.rows
        return self.clients[row * self.columns : row * self.columns + self.columns]

    def get_column(self, column):
        """Get all clients in given column"""
        assert column < self.columns
        return [self.clients[i] for i in range(column, len(self.clients), self.columns)]

    def add_client(self, client: Window) -> None:  # type: ignore[override]
        """Add client to Layout.
        Note that for Matrix the clients are appended at end of list.
        If needed a new row in matrix is created"""
        return self.clients.append(client)

    def configure(self, client: Window, screen_rect: ScreenRect) -> None:
        if client not in self.clients:
            return
        idx = self.clients.index(client)
        row = idx // self.columns
        col = idx % self.columns
        column_size = int(math.ceil(len(self.clients) / self.columns))
        if client.has_focus:
            px = self.border_focus
        else:
            px = self.border_normal
        # calculate position and size
        column_width = int(screen_rect.width / float(self.columns))
        row_height = int(screen_rect.height / float(column_size))
        xoffset = screen_rect.x + col * column_width
        yoffset = screen_rect.y + row * row_height
        win_width = column_width - 2 * self.border_width
        win_height = row_height - 2 * self.border_width
        # place
        client.place(
            xoffset,
            yoffset,
            win_width,
            win_height,
            self.border_width,
            px,
            margin=self.margin,
        )
        client.unhide()

    @expose_command()
    def previous(self) -> None:
        _SimpleLayoutBase.previous(self)

    @expose_command()
    def next(self) -> None:
        _SimpleLayoutBase.next(self)

    def horizontal_traversal(self, direction):
        """
        Internal method for determining left or right client.
        Negative direction is to left
        """
        column, row = self.column, self.row
        column = (column + direction) % len(self.get_row(row))
        self.clients.current_index = row * self.columns + column
        self.group.focus(self.clients.current_client)

    def vertical_traversal(self, direction):
        """
        internal method for determining above or below client.
        Negative direction is to top
        """
        column, row = self.column, self.row
        row = (row + direction) % len(self.get_column(column))
        self.clients.current_index = row * self.columns + column
        self.group.focus(self.clients.current_client)

    @expose_command()
    def left(self):
        """Switch to the next window on current row"""
        self.horizontal_traversal(-1)

    @expose_command()
    def right(self):
        """Switch to the next window on current row"""
        self.horizontal_traversal(+1)

    @expose_command()
    def up(self):
        """Switch to the previous window in current column"""
        self.vertical_traversal(-1)

    @expose_command()
    def down(self):
        """Switch to the next window in current column"""
        self.vertical_traversal(+1)

    @expose_command()
    def delete(self):
        """Decrease number of columns"""
        self.columns -= 1
        self.group.layout_all()

    @expose_command()
    def add(self):
        """Increase number of columns"""
        self.columns += 1
        self.group.layout_all()



import atexit
import enum
import os
import sys

from libqtile.log_utils import logger

__all__ = [
    "lifecycle",
]

Behavior = enum.Enum("Behavior", "NONE TERMINATE RESTART")


class LifeCycle:
    # This class exists mostly to move os.execv to the absolute last thing that
    # the python VM does before termination.
    # Be very careful about what references this class owns. Any object
    # referenced here when atexit fires will NOT be finalized properly.
    def __init__(self) -> None:
        self.behavior = Behavior.NONE
        self.exitcode: int = 0
        self.state_file: str | None = None
        atexit.register(self._atexit)

    def _atexit(self) -> None:
        if self.behavior is Behavior.RESTART:
            argv = sys.argv
            if "--no-spawn" not in argv:
                argv.append("--no-spawn")
            argv = [s for s in argv if not s.startswith("--with-state")]
            if self.state_file is not None:
                argv.append("--with-state=" + self.state_file)
            logger.warning("Restarting Qtile with os.execv(...)")
            # No other code will execute after the following line does
            os.execv(argv[0], argv)
        elif self.behavior is Behavior.TERMINATE:
            logger.warning("Qtile will now terminate")
            # the if statement should be unnecessary, but keeps code coverage working
            # calling os._exit prevents later injected atexit handlers from running
            if self.exitcode:
                os._exit(self.exitcode)
        elif self.behavior is Behavior.NONE:
            pass


lifecycle = LifeCycle()



from typing import Any

from libqtile import bar, hook
from libqtile.command.base import expose_command
from libqtile.layout import plasma
from libqtile.layout.plasma import AddMode
from libqtile.widget import base


class Plasma(base._TextBox):
    """
    A simple widget to indicate in which direction new windows will be
    added in the Plasma layout.
    """

    defaults: list[tuple[str, Any, str]] = [
        ("horizontal", "H", "Text to display if horizontal mode"),
        ("vertical", "V", "Text to display if horizontal mode"),
        ("split", "S", "Text to append to mode if ``horizontal/vertical_split`` not set"),
        ("horizontal_split", None, "Text to display for horizontal split mode"),
        ("vertical_split", None, "Text to display for horizontal split mode"),
        ("format", "{mode:>2}", "Format appearance of text"),
    ]

    def __init__(self, text="", width=bar.CALCULATED, **config):
        base._TextBox.__init__(self, text=text, width=width, **config)
        self.add_defaults(Plasma.defaults)
        self.add_callbacks({"Button1": self.next_mode})

        if self.horizontal_split is None:
            self.horizontal_split = self.horizontal + self.split

        if self.vertical_split is None:
            self.vertical_split = self.vertical + self.split

        self.modes = [
            AddMode.HORIZONTAL,
            AddMode.HORIZONTAL | AddMode.SPLIT,
            AddMode.VERTICAL,
            AddMode.VERTICAL | AddMode.SPLIT,
        ]
        self._mode = self.modes[0]
        self._layout = None

    def _configure(self, qtile, bar):
        base._TextBox._configure(self, qtile, bar)
        hook.subscribe.plasma_add_mode(self.mode_changed)
        hook.subscribe.layout_change(self.layout_changed)
        self.mode_changed(bar.screen.group.layout)

    def mode_changed(self, layout):
        """Update text depending on add_mode of layout."""
        if not isinstance(layout, plasma.Plasma):
            self.update("")
            self._layout = None
            return

        self._layout = layout

        if layout.group and layout.group.screen is not self.bar.screen:
            return

        if layout.horizontal:
            if layout.split:
                mode = self.horizontal_split
                self._mode = AddMode.HORIZONTAL | AddMode.SPLIT
            else:
                mode = self.horizontal
                self._mode = AddMode.HORIZONTAL
        else:
            if layout.split:
                mode = self.vertical_split
                self._mode = AddMode.VERTICAL | AddMode.SPLIT
            else:
                mode = self.vertical
                self._mode = AddMode.VERTICAL

        self.update(self.format.format(mode=mode))

    def layout_changed(self, layout, group):
        """Update widget when layout changes."""
        if group.screen is self.bar.screen:
            self.mode_changed(layout)

    def finalize(self):
        hook.unsubscribe.plasma_add_mode(self.mode_changed)
        hook.unsubscribe.layout_change(self.layout_changed)
        base._TextBox.finalize(self)

    @expose_command()
    def next_mode(self):
        """Change the add mode for the Plasma layout."""
        if self._layout is None:
            return

        index = self.modes.index(self._mode)
        index = (index + 1) % len(self.modes)

        self._layout.add_mode = self.modes[index]



from libqtile import bar, hook
from libqtile.widget import base


class Chord(base._TextBox):
    """Display current key chord"""

    defaults = [
        (
            "chords_colors",
            {},
            "colors per chord in form of tuple {'chord_name': ('bg', 'fg')}. "
            "Where a chord name is not in the dictionary, the default ``background`` and ``foreground``"
            " values will be used.",
        ),
        (
            "name_transform",
            lambda txt: txt,
            "preprocessor for chord name it is pure function string -> string",
        ),
    ]

    def __init__(self, width=bar.CALCULATED, **config):
        base._TextBox.__init__(self, "", width, **config)
        self.add_defaults(Chord.defaults)

    def _configure(self, qtile, bar):
        base._TextBox._configure(self, qtile, bar)
        self.default_background = self.background
        self.default_foreground = self.foreground
        self.text = ""
        self._setup_hooks()

    def _setup_hooks(self):
        def hook_enter_chord(chord_name):
            if chord_name is True:
                self.text = ""
                self.reset_colours()
                return

            self.text = self.name_transform(chord_name)
            if chord_name in self.chords_colors:
                self.background, self.layout.colour = self.chords_colors.get(chord_name)
            else:
                self.reset_colours()

            self.bar.draw()

        hook.subscribe.enter_chord(hook_enter_chord)
        hook.subscribe.leave_chord(self.clear)

    def reset_colours(self):
        self.background = self.default_background
        self.layout.colour = self.default_foreground

    def clear(self, *args):
        self.reset_colours()
        self.text = ""
        self.bar.draw()



from libqtile import DynamicLibraries, find_library
from libqtile.pango_ffi import pango_ffi as ffi

gobject = ffi.dlopen(find_library(DynamicLibraries.GOBJECT))  # type: ignore
pango = ffi.dlopen(find_library(DynamicLibraries.PANGO))  # type: ignore
pangocairo = ffi.dlopen(find_library(DynamicLibraries.PANGOCAIRO))  # type: ignore


def create_layout(cairo_t):
    """Create a PangoLayout from a cairo context."""
    return PangoLayout(cairo_t._pointer)


def show_layout(cairo_t, layout):
    """Show a PangoLayout on a cairo context."""
    pangocairo.pango_cairo_show_layout(cairo_t._pointer, layout._pointer)


ALIGN_CENTER = pango.PANGO_ALIGN_CENTER
ELLIPSIZE_END = pango.PANGO_ELLIPSIZE_END
units_from_double = pango.pango_units_from_double
units_to_double = pango.pango_units_to_double


ALIGNMENTS = {
    "left": pango.PANGO_ALIGN_LEFT,
    "center": pango.PANGO_ALIGN_CENTER,
    "right": pango.PANGO_ALIGN_RIGHT,
}


class PangoLayout:
    def __init__(self, cairo_t):
        self._cairo_t = cairo_t
        self._pointer = pangocairo.pango_cairo_create_layout(cairo_t)

        def free(p):
            gobject.g_object_unref(p)

        self._pointer = ffi.gc(self._pointer, free)

    def finalize(self):
        self._desc = None
        self._pointer = None
        self._cairo_t = None

    def set_font_description(self, desc):
        # save a pointer so it doesn't get GC'd out from under us
        self._desc = desc
        pango.pango_layout_set_font_description(self._pointer, desc._pointer)

    def get_font_description(self):
        descr = pango.pango_layout_get_font_description(self._pointer)
        return FontDescription(descr)

    def set_alignment(self, alignment):
        pango.pango_layout_set_alignment(self._pointer, alignment)

    def set_attributes(self, attrs):
        pango.pango_layout_set_attributes(self._pointer, attrs)

    def set_text(self, text):
        text = text.encode("utf-8")
        pango.pango_layout_set_text(self._pointer, text, -1)

    def get_text(self):
        ret = pango.pango_layout_get_text(self._pointer)
        return ffi.string(ret).decode()

    def set_ellipsize(self, ellipzize):
        pango.pango_layout_set_ellipsize(self._pointer, ellipzize)

    def get_ellipsize(self):
        return pango.pango_layout_get_ellipsize(self._pointer)

    def get_pixel_size(self):
        width = ffi.new("int[1]")
        height = ffi.new("int[1]")

        pango.pango_layout_get_pixel_size(self._pointer, width, height)

        return width[0], height[0]

    def set_width(self, width):
        pango.pango_layout_set_width(self._pointer, width)


class FontDescription:
    def __init__(self, pointer=None):
        if pointer is None:
            self._pointer = pango.pango_font_description_new()
            self._pointer = ffi.gc(self._pointer, pango.pango_font_description_free)
        else:
            self._pointer = pointer

    @classmethod
    def from_string(cls, string):
        pointer = pango.pango_font_description_from_string(string.encode())
        pointer = ffi.gc(pointer, pango.pango_font_description_free)
        return cls(pointer)

    def set_family(self, family):
        pango.pango_font_description_set_family(self._pointer, family.encode())

    def get_family(self):
        ret = pango.pango_font_description_get_family(self._pointer)
        return ffi.string(ret).decode()

    def set_absolute_size(self, size):
        pango.pango_font_description_set_absolute_size(self._pointer, size)

    def set_size(self, size):
        pango.pango_font_description_set_size(self._pointer, size)

    def get_size(self):
        return pango.pango_font_description_get_size(self._pointer)


class BadMarkup(Exception):
    pass


def parse_markup(value, accel_marker=0):
    attr_list = ffi.new("PangoAttrList**")
    text = ffi.new("char**")
    error = ffi.new("GError**")
    markup_text = value.encode()

    ret = pango.pango_parse_markup(
        markup_text, -1, accel_marker, attr_list, text, ffi.NULL, error
    )

    if ret == 0:
        raise BadMarkup(f"parse_markup() failed for: {value}")

    return attr_list[0], ffi.string(text[0]), chr(accel_marker)


def markup_escape_text(text):
    ret = gobject.g_markup_escape_text(text.encode("utf-8"), -1)
    return ffi.string(ret).decode()



from itertools import cycle

from psutil import getloadavg

from libqtile.command.base import expose_command
from libqtile.widget import base


class Load(base.InLoopPollText):
    """
    A small widget to show the load averages of the system.
    Depends on psutil.
    """

    defaults = [
        ("update_interval", 1.0, "The update interval for the widget"),
        ("format", "Load({time}):{load:.2f}", "The format in which to display the results."),
    ]
    times = ["1m", "5m", "15m"]

    def __init__(self, **config):
        super().__init__("", **config)
        self.add_defaults(Load.defaults)
        self.add_callbacks({"Button1": self.next_load})
        self.cycled_times = cycle(Load.times)
        self.set_time()

    def set_time(self):
        self.time = next(self.cycled_times)

    @expose_command()
    def next_load(self):
        self.set_time()
        self.update(self.poll())

    def poll(self):
        loads = {}
        (
            loads["1m"],
            loads["5m"],
            loads["15m"],
        ) = getloadavg()  # Gets the load averages as a dictionary.
        load = loads[self.time]
        return self.format.format(time=self.time, load=load)



__all__ = [
    "Bsp",
    "Columns",
    "Floating",
    "Matrix",
    "Max",
    "Plasma",
    "RatioTile",
    "ScreenSplit",
    "Slice",
    "Spiral",
    "Stack",
    "Tile",
    "TreeTab",
    "VerticalTile",
    "MonadTall",
    "MonadThreeCol",
    "MonadWide",
    "Zoomy",
]

from libqtile.layout.bsp import Bsp
from libqtile.layout.columns import Columns
from libqtile.layout.floating import Floating
from libqtile.layout.matrix import Matrix
from libqtile.layout.max import Max
from libqtile.layout.plasma import Plasma
from libqtile.layout.ratiotile import RatioTile
from libqtile.layout.screensplit import ScreenSplit
from libqtile.layout.slice import Slice
from libqtile.layout.spiral import Spiral
from libqtile.layout.stack import Stack
from libqtile.layout.tile import Tile
from libqtile.layout.tree import TreeTab
from libqtile.layout.verticaltile import VerticalTile
from libqtile.layout.xmonad import MonadTall, MonadThreeCol, MonadWide
from libqtile.layout.zoomy import Zoomy



import datetime

from libqtile.widget.gen_poll_url import GenPollUrl


class IdleRPG(GenPollUrl):
    """
    A widget for monitoring and displaying IdleRPG stats.

    ::

        # display idlerpg stats for the player 'pants' on freenode's #idlerpg
        widget.IdleRPG(url="http://xethron.lolhosting.net/xml.php?player=pants")

    Widget requirements: xmltodict_.

    .. _xmltodict: https://pypi.org/project/xmltodict/
    """

    defaults = [
        ("format", "IdleRPG: {online} TTL: {ttl}", "Display format"),
    ]

    def __init__(self, **config):
        GenPollUrl.__init__(self, **config)
        self.add_defaults(IdleRPG.defaults)
        self.json = False
        self.xml = True

    def parse(self, body):
        formatted = {}
        for k, v in body["player"].items():
            if k == "ttl":
                formatted[k] = str(datetime.timedelta(seconds=int(v)))
            elif k == "online":
                formatted[k] = "online" if v == "1" else "offline"
            else:
                formatted[k] = v

        return self.format.format(**formatted)



"""
Command-line wrapper to run commands and add rules to new windows
"""

import argparse
import atexit
import subprocess

from libqtile import ipc
from libqtile.command import graph


def run_cmd(opts) -> None:
    if opts.socket is None:
        socket = ipc.find_sockfile()
    else:
        socket = opts.socket
    client = ipc.Client(socket)
    root = graph.CommandGraphRoot()

    cmd = [opts.cmd]
    if opts.args:
        cmd.extend(opts.args)

    proc = subprocess.Popen(cmd)
    match_args = {"net_wm_pid": proc.pid}
    rule_args = {
        "float": opts.float,
        "intrusive": opts.intrusive,
        "group": opts.group,
        "break_on_match": not opts.dont_break,
    }

    graph_cmd = root.call("add_rule")
    # client.send(selectors, name, args, kwargs, lifted)
    _, rule_id = client.send((root.selectors, graph_cmd.name, (match_args, rule_args), {}, False))

    def remove_rule() -> None:
        cmd = root.call("remove_rule")
        # client.send(selectors, name, args, kwargs, lifted)
        client.send((root.selectors, cmd.name, (rule_id,), {}, False))

    atexit.register(remove_rule)

    proc.wait()


def add_subcommand(subparsers, parents):
    parser = subparsers.add_parser(
        "run-cmd", parents=parents, help="A wrapper around the command graph."
    )
    parser.add_argument("-s", "--socket", help="Use specified socket for IPC.")
    parser.add_argument(
        "-i", "--intrusive", action="store_true", help="If the new window should be intrusive."
    )
    parser.add_argument(
        "-f", "--float", action="store_true", help="If the new window should be floating."
    )
    parser.add_argument(
        "-b",
        "--dont-break",
        action="store_true",
        help="Do not break on match (keep applying rules).",
    )
    parser.add_argument("-g", "--group", help="Set the window group.")
    parser.add_argument("cmd", help="Command to execute.")
    parser.add_argument(
        "args",
        nargs=argparse.REMAINDER,
        metavar="[args ...]",
        help="Optional arguments to pass to command.",
    )
    parser.set_defaults(func=run_cmd)



from libqtile.backend.base.core import Core  # noqa: F401
from libqtile.backend.base.drawer import Drawer  # noqa: F401
from libqtile.backend.base.float_states import FloatStates  # noqa: F401
from libqtile.backend.base.window import (  # noqa: F401
    Internal,
    Static,
    Window,
    WindowType,
)



import collections

import libqtile.hook
from libqtile.backend.base import Static
from libqtile.config import Group, Key, Rule
from libqtile.lazy import lazy
from libqtile.log_utils import logger


def simple_key_binder(mod, keynames=None):
    """Bind keys to mod+group position or to the keys specified as second argument"""

    def func(dgroup):
        # unbind all
        for key in dgroup.keys[:]:
            dgroup.qtile.ungrab_key(key)
            dgroup.qtile.config.keys.remove(key)
            dgroup.keys.remove(key)

        if keynames:
            keys = keynames
        else:
            # keys 1 to 9 and 0
            keys = list(map(str, list(range(1, 10)) + [0]))

        # bind all keys
        for keyname, group in zip(keys, dgroup.qtile.groups):
            name = group.name
            key = Key([mod], keyname, lazy.group[name].toscreen())
            key_s = Key([mod, "shift"], keyname, lazy.window.togroup(name))
            key_c = Key([mod, "control"], keyname, lazy.group.switch_groups(name))
            dgroup.keys.extend([key, key_s, key_c])
            dgroup.qtile.config.keys.extend([key, key_s, key_c])
            dgroup.qtile.grab_key(key)
            dgroup.qtile.grab_key(key_s)
            dgroup.qtile.grab_key(key_c)

    return func


class DGroups:
    """Dynamic Groups"""

    def __init__(self, qtile, dgroups, key_binder=None, delay=1):
        self.qtile = qtile

        self.groups = dgroups
        self.groups_map = {}

        self.rules = []
        self.rules_map = {}
        self.last_rule_id = 0

        for rule in getattr(qtile.config, "dgroups_app_rules", []):
            self.add_rule(rule)

        self.keys = []

        self.key_binder = key_binder

        self._setup_hooks()
        self._setup_groups()

        self.delay = delay

        self.timeout = {}

    def add_rule(self, rule, last=True):
        rule_id = self.last_rule_id
        self.rules_map[rule_id] = rule
        if last:
            self.rules.append(rule)
        else:
            self.rules.insert(0, rule)
        self.last_rule_id += 1
        return rule_id

    def remove_rule(self, rule_id):
        rule = self.rules_map.get(rule_id)
        if rule:
            self.rules.remove(rule)
            del self.rules_map[rule_id]
        else:
            logger.warning('Rule "%s" not found', rule_id)

    def add_dgroup(self, group, start=False):
        self.groups_map[group.name] = group
        rule = Rule(group.matches, group=group.name)
        self.rules.append(rule)
        if start:
            self.qtile.add_group(
                group.name,
                group.layout,
                group.layouts,
                group.label,
                screen_affinity=group.screen_affinity,
            )

    def _setup_groups(self):
        def launch_app(app, group):
            def launcher():
                self.qtile.spawn(app, group=group)

            return launcher

        for group in self.groups:
            self.add_dgroup(group, group.init)

            if group.spawn and not self.qtile.no_spawn:
                if isinstance(group.spawn, str):
                    spawns = [group.spawn]
                else:
                    spawns = group.spawn
                for spawn in spawns:
                    libqtile.hook.subscribe.startup_once(launch_app(spawn, group=group.name))

    def _setup_hooks(self):
        libqtile.hook.subscribe.addgroup(self._addgroup)
        libqtile.hook.subscribe.client_new(self._add)
        libqtile.hook.subscribe.client_killed(self._del)
        if self.key_binder:
            libqtile.hook.subscribe.setgroup(lambda: self.key_binder(self))
            libqtile.hook.subscribe.changegroup(lambda: self.key_binder(self))

    def _addgroup(self, group_name):
        if group_name not in self.groups_map:
            self.add_dgroup(Group(group_name, persist=False))

    def _add(self, client):
        if client in self.timeout:
            logger.debug("Remove dgroup source")
            self.timeout.pop(client).cancel()

        # ignore static windows
        if isinstance(client, Static):
            return

        # ignore windows whose groups is already set (e.g. from another hook or
        # when it was set on state restore)
        if client.group is not None:
            return

        group_set = False
        intrusive = False

        delete_rules = []

        for rule in self.rules:
            # Matching Rules
            if rule.matches(client):
                if rule.group:
                    if rule.group in self.groups_map:
                        layout = self.groups_map[rule.group].layout
                        layouts = self.groups_map[rule.group].layouts
                        label = self.groups_map[rule.group].label
                    else:
                        layout = None
                        layouts = None
                        label = None
                    group_added = self.qtile.add_group(rule.group, layout, layouts, label)
                    client.togroup(rule.group)

                    group_set = True

                    group_obj = self.qtile.groups_map[rule.group]
                    group = self.groups_map.get(rule.group)
                    if group and group_added:
                        for k, v in list(group.layout_opts.items()):
                            if isinstance(v, collections.abc.Callable):
                                v(group_obj.layout)
                            else:
                                setattr(group_obj.layout, k, v)
                        affinity = group.screen_affinity
                        if affinity and len(self.qtile.screens) > affinity:
                            self.qtile.screens[affinity].set_group(group_obj)

                if rule.float:
                    client.enable_floating()

                if rule.intrusive:
                    intrusive = rule.intrusive

                if rule.one_time:
                    delete_rules.append(rule)

                if rule.break_on_match:
                    break

        if delete_rules:
            ids_to_delete = [
                rule_id for rule_id, rule in self.rules_map.items() if rule in delete_rules
            ]
            for rule_id in ids_to_delete:
                self.remove_rule(rule_id)

        # If app doesn't have a group
        if not group_set:
            current_group = self.qtile.current_group.name
            if (
                current_group in self.groups_map
                and self.groups_map[current_group].exclusive
                and not intrusive
            ):
                wm_class = client.get_wm_class()

                if wm_class:
                    if len(wm_class) > 1:
                        wm_class = wm_class[1]
                    else:
                        wm_class = wm_class[0]

                    group_name = wm_class
                else:
                    group_name = client.name or "Unnamed"

                self.add_dgroup(Group(group_name, persist=False), start=True)
                client.togroup(group_name)
        self.sort_groups()

    def sort_groups(self):
        grps = self.qtile.groups
        sorted_grps = sorted(grps, key=lambda g: self.groups_map[g.name].position)
        if grps != sorted_grps:
            self.qtile.groups = sorted_grps
            libqtile.hook.fire("changegroup")

    def _del(self, client):
        # ignore static windows
        if isinstance(client, Static):
            return

        group = client.group

        def delete_client():
            # Delete group if empty and don't persist
            if (
                group
                and group.name in self.groups_map
                and not self.groups_map[group.name].persist
                and len(group.windows) <= 0
            ):
                self.qtile.delete_group(group.name)
                self.sort_groups()
            del self.timeout[client]

        if group is not None and group.persist:
            return

        logger.debug("Deleting %s in %ss", group, self.delay)
        if client not in self.timeout:
            self.timeout[client] = self.qtile.call_later(self.delay, delete_client)



from __future__ import annotations

import locale
from os import makedirs, path
from pathlib import Path
from sys import exit
from typing import TYPE_CHECKING

import libqtile.backend
from libqtile import confreader, qtile
from libqtile.log_utils import logger
from libqtile.utils import VERSION, get_config_file

if TYPE_CHECKING:
    from libqtile.core.manager import Qtile


LIBQTILE_PATH = Path(__file__).parent.parent


def rename_process():
    """
    Try to rename the qtile process if py-setproctitle is installed:

    http://code.google.com/p/py-setproctitle/

    Will fail silently if it's not installed. Setting the title lets you do
    stuff like "killall qtile".
    """
    try:
        import setproctitle

        setproctitle.setproctitle("qtile")
    except ImportError:
        pass


def get_default_config():
    config = LIBQTILE_PATH / "resources" / "default_config.py"
    return config.resolve().as_posix()


def make_qtile(options) -> Qtile | None:
    qtile.core.name = options.backend
    if missing_deps := libqtile.backend.has_deps(options.backend):
        print(f"Backend '{options.backend}' missing required Python dependencies:")
        for dep in missing_deps:
            print("\t", dep)

        return None

    kore = libqtile.backend.get_core(options.backend)

    if not path.isfile(options.configfile):
        try:
            makedirs(path.dirname(options.configfile), exist_ok=True)
            from shutil import copyfile

            default_config_path = path.join(
                path.dirname(__file__), "..", "resources", "default_config.py"
            )
            copyfile(default_config_path, options.configfile)
            logger.info("Copied default_config.py to %s", options.configfile)
        except Exception:
            logger.exception("Failed to copy default_config.py to %s:", options.configfile)

    config = confreader.Config(options.configfile)

    # XXX: the import is here because we need to call init_log
    # before start importing stuff
    from libqtile.core.manager import Qtile

    return Qtile(
        kore,
        config,
        no_spawn=options.no_spawn,
        state=options.state,
        socket_path=options.socket,
    )


def start(options):
    try:
        locale.setlocale(locale.LC_ALL, "")
    except locale.Error:
        pass

    libpath = (LIBQTILE_PATH / "..").resolve()
    logger.warning(f"Starting Qtile {VERSION} from {libpath}")
    rename_process()
    q = make_qtile(options)
    if q is None:
        logger.warning("Backend is missing required Python dependencies. Exiting.")
        exit(1)

    try:
        q.loop()
    except Exception:
        logger.exception("Qtile crashed")
        exit(1)
    logger.info("Exiting...")


def add_subcommand(subparsers, parents):
    parser = subparsers.add_parser("start", parents=parents, help="Start a Qtile session.")
    configs = parser.add_mutually_exclusive_group()
    configs.add_argument(
        "-c",
        "--config",
        action="store",
        default=get_config_file(),
        dest="configfile",
        help="Use the specified configuration file.",
    )
    configs.add_argument(
        "-d",
        "--use-default-config",
        action="store_const",
        const=get_default_config(),
        dest="configfile",
        help="Use qtile's default configuration file.",
    )
    parser.add_argument(
        "-s",
        "--socket",
        action="store",
        default=None,
        dest="socket",
        help="Use specified socket for IPC.",
    )
    parser.add_argument(
        "-n",
        "--no-spawn",
        action="store_true",
        default=False,
        dest="no_spawn",
        help="Avoid spawning apps. (Used when restarting Qtile)",
    )
    parser.add_argument(
        "--with-state",
        default=None,
        dest="state",
        help="Pickled QtileState object. (Used when restarting Qtile).",
    )
    parser.add_argument(
        "-b",
        "--backend",
        default="x11",
        dest="backend",
        choices=libqtile.backend.CORES.keys(),
        help="Use specified backend.",
    )
    parser.set_defaults(func=start)



from __future__ import annotations

import importlib
import importlib.util
from typing import TYPE_CHECKING

from libqtile.utils import QtileError

if TYPE_CHECKING:
    from typing import Any

    from libqtile.backend.base import Core


CORES = {
    "wayland": (),
    "x11": ("xcffib",),
}


def has_deps(backend: str) -> list[str]:
    """
    Check if the backend has all its dependencies installed.

    Args:
        backend: The backend to check.

    Returns:
        A list of missing dependencies. If this is empty, we can use this backend.
    """
    if backend not in CORES:
        raise QtileError(f"Backend {backend} does not exist")

    not_found = []
    for dep in CORES[backend]:
        if not importlib.util.find_spec(dep):
            not_found.append(dep)

    return not_found


def get_core(backend: str, *args: Any) -> Core:
    if backend not in CORES:
        raise QtileError(f"Backend {backend} does not exist")

    return importlib.import_module(f"libqtile.backend.{backend}.core").Core(*args)



from datetime import datetime, timedelta
from time import time

from libqtile.command.base import expose_command
from libqtile.utils import send_notification
from libqtile.widget import base


class Pomodoro(base.InLoopPollText):
    """Pomodoro technique widget"""

    defaults = [
        ("num_pomodori", 4, "Number of pomodori to do in a cycle"),
        ("length_pomodori", 25, "Length of one pomodori in minutes"),
        ("length_short_break", 5, "Length of a short break in minutes"),
        ("length_long_break", 15, "Length of a long break in minutes"),
        ("color_inactive", "ff0000", "Colour then pomodoro is inactive"),
        ("color_active", "00ff00", "Colour then pomodoro is running"),
        ("color_break", "ffff00", "Colour then it is break time"),
        ("notification_on", True, "Turn notifications on"),
        ("prefix_inactive", "POMODORO", "Prefix when app is inactive"),
        ("prefix_active", "", "Prefix then app is active"),
        ("prefix_break", "B ", "Prefix during short break"),
        ("prefix_long_break", "LB ", "Prefix during long break"),
        ("prefix_paused", "PAUSE", "Prefix during pause"),
        (
            "update_interval",
            1,
            "Update interval in seconds, if none, the "
            "widget updates whenever the event loop is idle.",
        ),
    ]

    STATUS_START = "start"
    STATUS_INACTIVE = "inactive"
    STATUS_ACTIVE = "active"
    STATUS_BREAK = "break"
    STATUS_LONG_BREAK = "long_break"
    STATUS_PAUSED = "paused"

    status = "inactive"
    paused_status = None
    end_time = datetime.now()
    time_left = None
    pomodoros = 1

    def __init__(self, **config):
        base.InLoopPollText.__init__(self, "", **config)
        self.add_defaults(Pomodoro.defaults)
        self.prefix = {
            "inactive": self.prefix_inactive,
            "active": self.prefix_active,
            "break": self.prefix_break,
            "long_break": self.prefix_long_break,
            "paused": self.prefix_paused,
        }

        self.add_callbacks(
            {
                "Button1": self.toggle_break,
                "Button3": self.toggle_active,
            }
        )

    def tick(self):
        self.update(self.poll())
        return self.update_interval - time() % self.update_interval

    def _update(self):
        if self.status in [self.STATUS_INACTIVE, self.STATUS_PAUSED]:
            return

        if self.end_time > datetime.now() and self.status != self.STATUS_START:
            return

        if self.status == self.STATUS_ACTIVE and self.pomodoros == self.num_pomodori:
            self.status = self.STATUS_LONG_BREAK
            self.end_time = datetime.now() + timedelta(minutes=self.length_long_break)
            self.pomodoros = 1
            if self.notification_on:
                self._send_notification(
                    "normal",
                    "Please take a long break! End Time: " + self.end_time.strftime("%H:%M"),
                )
            return

        if self.status == self.STATUS_ACTIVE:
            self.status = self.STATUS_BREAK
            self.end_time = datetime.now() + timedelta(minutes=self.length_short_break)
            self.pomodoros += 1
            if self.notification_on:
                self._send_notification(
                    "normal",
                    "Please take a short break! End Time: " + self.end_time.strftime("%H:%M"),
                )
            return

        self.status = self.STATUS_ACTIVE
        self.end_time = datetime.now() + timedelta(minutes=self.length_pomodori)
        if self.notification_on:
            self._send_notification(
                "critical",
                "Please start with the next Pomodori! End Time: "
                + self.end_time.strftime("%H:%M"),
            )

        return

    def _get_text(self):
        self._update()

        if self.status in [self.STATUS_INACTIVE, self.STATUS_PAUSED]:
            self.layout.colour = self.color_inactive
            return self.prefix[self.status]

        time_left = self.end_time - datetime.now()

        if self.status == self.STATUS_ACTIVE:
            self.layout.colour = self.color_active
        else:
            self.layout.colour = self.color_break

        time_string = f"{time_left.seconds // 3600:d}:{time_left.seconds % 3600 // 60:02d}:{time_left.seconds % 60:02d}"
        return self.prefix[self.status] + time_string

    @expose_command()
    def toggle_break(self):
        if self.status == self.STATUS_INACTIVE:
            self.status = self.STATUS_START
            return

        if self.paused_status is None:
            self.paused_status = self.status
            self.time_left = self.end_time - datetime.now()
            self.status = self.STATUS_PAUSED
            if self.notification_on:
                self._send_notification("low", "Pomodoro has been paused")
        else:
            self.status = self.paused_status
            self.paused_status = None
            self.end_time = self.time_left + datetime.now()
            if self.notification_on:
                if self.status == self.STATUS_ACTIVE:
                    status = "Pomodoro"
                else:
                    status = "break"

                self._send_notification(
                    "normal",
                    f"Please continue on {status}! End Time: " + self.end_time.strftime("%H:%M"),
                )

    @expose_command()
    def toggle_active(self):
        if self.status != self.STATUS_INACTIVE:
            self.status = self.STATUS_INACTIVE
            if self.notification_on:
                self._send_notification("critical", "Pomodoro has been suspended")
        else:
            self.status = self.STATUS_START

    def _send_notification(self, urgent, message):
        send_notification("Pomodoro", message, urgent=urgent)

    def poll(self):
        return self._get_text()



from __future__ import annotations

import typing
from abc import ABCMeta, abstractmethod

from libqtile import hook
from libqtile.command.base import CommandError, CommandObject, expose_command
from libqtile.log_utils import logger
from libqtile.scratchpad import ScratchPad

if typing.TYPE_CHECKING:
    from typing import Any

    from libqtile import config
    from libqtile.backend.base import Drawer
    from libqtile.command.base import ItemT
    from libqtile.core.manager import Qtile
    from libqtile.group import _Group
    from libqtile.utils import ColorsType


class _Window(CommandObject, metaclass=ABCMeta):
    def __init__(self):
        self.borderwidth: int = 0
        self.name: str = "<no name>"
        self.reserved_space: tuple[int, int, int, int] | None = None
        # Window.static sets this in case it is hooked to client_new to stop the
        # Window object from being managed, now that a Static is being used instead
        self.defunct: bool = False
        self._can_steal_focus: bool = True

        self.base_x: int | None = None
        self.base_y: int | None = None
        self.base_width: int | None = None
        self.base_height: int | None = None

    @property
    @abstractmethod
    def wid(self) -> int:
        """The unique window ID"""

    @abstractmethod
    def hide(self) -> None:
        """Hide the window"""

    @abstractmethod
    def unhide(self) -> None:
        """Unhide the window"""

    @expose_command()
    def is_visible(self) -> bool:
        """Is this window visible (i.e. not hidden)?"""
        return False

    @abstractmethod
    @expose_command()
    def kill(self) -> None:
        """Kill the window"""

    def get_wm_class(self) -> list | None:
        """Return the class(es) of the window"""
        return None

    def get_wm_type(self) -> str | None:
        """Return the type of the window"""
        return None

    def get_wm_role(self) -> str | None:
        """Return the role of the window"""
        return None

    @property
    def can_steal_focus(self) -> bool:
        """Is it OK for this window to steal focus?"""
        return self._can_steal_focus

    @can_steal_focus.setter
    def can_steal_focus(self, can_steal_focus: bool) -> None:
        """Can_steal_focus setter."""
        self._can_steal_focus = can_steal_focus

    def has_fixed_ratio(self) -> bool:
        """Does this window want a fixed aspect ratio?"""
        return False

    def has_fixed_size(self) -> bool:
        """Does this window want a fixed size?"""
        return False

    @property
    def urgent(self) -> bool:
        """Whether this window urgently wants focus"""
        return False

    @urgent.setter
    def urgent(self, urgent: bool) -> None:
        raise NotImplementedError

    @property
    def opacity(self) -> float:
        """The opacity of this window from 0 (transparent) to 1 (opaque)."""
        return self._opacity

    @opacity.setter
    def opacity(self, opacity: float) -> None:
        """Opacity setter."""
        self._opacity = opacity

    @abstractmethod
    @expose_command()
    def place(
        self,
        x,
        y,
        width,
        height,
        borderwidth,
        bordercolor,
        above=False,
        margin=None,
        respect_hints=False,
    ):
        """Place the window in the given position."""

    def _items(self, name: str) -> ItemT:
        return None

    def _select(self, name, sel):
        return None

    def _save_geometry(self):
        """Save current window geometry."""
        self.base_x = self.x
        self.base_y = self.y
        self.base_width = self.width
        self.base_height = self.height

    def _restore_geometry(self):
        """Restore previously saved window geometry."""
        if self.base_x is not None:
            self.x = self.base_x
        if self.base_y is not None:
            self.y = self.base_y
        if self.base_width is not None:
            self.width = self.base_width
        if self.base_height is not None:
            self.height = self.base_height

    @abstractmethod
    @expose_command()
    def info(self) -> dict[str, Any]:
        """
        Return information on this window.

        Mimimum required keys are:
        - name
        - x
        - y
        - width
        - height
        - group
        - id
        - wm_class

        """
        return {}

    @expose_command()
    def keep_above(self, enable: bool | None = None):
        """Keep this window above all others"""

    @expose_command()
    def keep_below(self, enable: bool | None = None):
        """Keep this window below all others"""

    @expose_command()
    def move_up(self, force: bool = False) -> None:
        """
        Move this window above the next window along the z axis.

        Will not raise a "normal" window (i.e. one that is not "kept_above/below")
        above a window that is marked as "kept_above".

        Will not raise a window where "keep_below" is True unless
        force is set to True.
        """

    @expose_command()
    def move_down(self, force: bool = False) -> None:
        """
        Move this window below the previous window along the z axis.

        Will not lower a "normal" window (i.e. one that is not "kept_above/below")
        below a window that is marked as "kept_below".

        Will not lower a window where "keep_above" is True unless
        force is set to True.
        """

    @expose_command()
    def move_to_top(self) -> None:
        """
        Move this window above all windows in the current layer
        e.g. if you have 3 windows all with "keep_above" set, calling
        this method will move the window to the top of those three windows.

        Calling this on a "normal" window will not raise it above a "kept_above"
        window.
        """

    @expose_command()
    def move_to_bottom(self) -> None:
        """
        Move this window below all windows in the current layer
        e.g. if you have 3 windows all with "keep_above" set, calling
        this method will move the window to the bottom of those three windows.

        Calling this on a "normal" window will not raise it below a "kept_below"
        window.
        """

    @abstractmethod
    @expose_command()
    def bring_to_front(self) -> None:
        """
        Bring the window to the front.

        In X11, `bring_to_front` ignores all other layering rules and brings the
        window to the very front. When that window loses focus, it will be stacked
        again according the appropriate rules.
        """


class Window(_Window, metaclass=ABCMeta):
    """
    A regular Window belonging to a client.

    Abstract methods are required to be defined as part of a specific backend's
    implementation. Non-abstract methods have default implementations here to be shared
    across backends.
    """

    qtile: Qtile

    # If float_x or float_y are None, the window has never been placed
    float_x: int | None
    float_y: int | None
    bordercolor: ColorsType | None

    def __repr__(self):
        return f"{self.__class__.__name__!s}(name={self.name!r}, wid={self.wid:d})"

    @property
    @abstractmethod
    def group(self) -> _Group | None:
        """The group to which this window belongs."""

    @group.setter
    def group(self, group: _Group | None) -> None:
        """Set the group."""

    @property
    def floating(self) -> bool:
        """Whether this window is floating."""
        return False

    @floating.setter
    def floating(self, do_float: bool) -> None:
        raise NotImplementedError

    @property
    def maximized(self) -> bool:
        """Whether this window is maximized."""
        return False

    @maximized.setter
    def maximized(self, do_maximize: bool) -> None:
        raise NotImplementedError

    @property
    def minimized(self) -> bool:
        """Whether this window is minimized."""
        return False

    @minimized.setter
    def minimized(self, do_minimize: bool) -> None:
        raise NotImplementedError

    @property
    def fullscreen(self) -> bool:
        """Whether this window is fullscreened."""
        return False

    @fullscreen.setter
    def fullscreen(self, do_full: bool) -> None:
        raise NotImplementedError

    @property
    def wants_to_fullscreen(self) -> bool:
        """Does this window want to be fullscreen?"""
        return False

    def match(self, match: config._Match) -> bool:
        """Compare this window against a Match instance."""
        return match.compare(self)

    @abstractmethod
    @expose_command()
    def focus(self, warp: bool = True) -> None:
        """Focus this window and optional warp the pointer to it."""

    @property
    def has_focus(self):
        return self == self.qtile.current_window

    def has_user_set_position(self) -> bool:
        """Whether this window has user-defined geometry"""
        return False

    def is_placed(self) -> bool:
        """Whether this window has been placed, i.e. both float offsets are not None."""
        return (
            self.group is not None
            and self.group.screen is not None
            and self.float_x is not None
            and self.float_y is not None
        )

    def is_transient_for(self) -> WindowType | None:
        """What window is this window a transient window for?"""
        return None

    @abstractmethod
    def get_pid(self) -> int:
        """Return the PID that owns the window."""

    def paint_borders(self, color: ColorsType, width: int) -> None:
        """Paint the window borders with the given color(s) and width"""

    @abstractmethod
    @expose_command()
    def get_position(self) -> tuple[int, int]:
        """Get the (x, y) of the window"""

    @abstractmethod
    @expose_command()
    def get_size(self) -> tuple[int, int]:
        """Get the (width, height) of the window"""

    @abstractmethod
    @expose_command()
    def move_floating(self, dx: int, dy: int) -> None:
        """Move window by dx and dy"""

    @abstractmethod
    @expose_command()
    def resize_floating(self, dw: int, dh: int) -> None:
        """Add dw and dh to size of window"""

    @abstractmethod
    @expose_command()
    def set_position_floating(self, x: int, y: int) -> None:
        """Move window to x and y"""

    @abstractmethod
    @expose_command()
    def set_position(self, x: int, y: int) -> None:
        """
        Move floating window to x and y; swap tiling window with the window under the
        pointer.
        """

    @abstractmethod
    @expose_command()
    def set_size_floating(self, w: int, h: int) -> None:
        """Set window dimensions to w and h"""

    @abstractmethod
    @expose_command()
    def toggle_floating(self) -> None:
        """Toggle the floating state of the window."""

    @abstractmethod
    @expose_command()
    def enable_floating(self) -> None:
        """Float the window."""

    @abstractmethod
    @expose_command()
    def disable_floating(self) -> None:
        """Tile the window."""

    @abstractmethod
    @expose_command()
    def toggle_maximize(self) -> None:
        """Toggle the fullscreen state of the window."""

    @abstractmethod
    @expose_command()
    def toggle_minimize(self) -> None:
        """Toggle the minimized state of the window."""

    @abstractmethod
    @expose_command()
    def toggle_fullscreen(self) -> None:
        """Toggle the fullscreen state of the window."""

    @abstractmethod
    @expose_command()
    def enable_fullscreen(self) -> None:
        """Fullscreen the window"""

    @abstractmethod
    @expose_command()
    def disable_fullscreen(self) -> None:
        """Un-fullscreen the window"""

    @abstractmethod
    @expose_command()
    def togroup(
        self, group_name: str | None = None, switch_group: bool = False, toggle: bool = False
    ) -> None:
        """Move window to a specified group

        Also switch to that group if `switch_group` is True.

        If `toggle` is True and and the specified group is already on the screen,
        use the last used group as target instead.

        Examples
        ========

        Move window to current group::

            togroup()

        Move window to group "a"::

            togroup("a")

        Move window to group "a", and switch to group "a"::

            togroup("a", switch_group=True)
        """
        self.togroup(group_name, switch_group=switch_group, toggle=toggle)

    @expose_command()
    def toscreen(self, index: int | None = None) -> None:
        """Move window to a specified screen.

        If index is not specified, we assume the current screen

        Examples
        ========

        Move window to current screen::

            toscreen()

        Move window to screen 0::

            toscreen(0)
        """
        if index is None:
            screen = self.qtile.current_screen
        else:
            try:
                screen = self.qtile.screens[index]
            except IndexError:
                raise CommandError(f"No such screen: {index:d}")
        self.togroup(screen.group.name)

    @expose_command()
    def set_opacity(self, opacity: float) -> None:
        """Set the window's opacity"""
        if opacity < 0.1:
            self.opacity = 0.1
        elif opacity > 1:
            self.opacity = 1
        else:
            self.opacity = opacity

    @expose_command()
    def down_opacity(self) -> None:
        """Decrease the window's opacity by 10%."""
        self.set_opacity(self.opacity - 0.1)

    @expose_command()
    def up_opacity(self) -> None:
        """Increase the window's opacity by 10%."""
        self.set_opacity(self.opacity + 0.1)

    @abstractmethod
    @expose_command()
    def static(
        self,
        screen: int | None = None,
        x: int | None = None,
        y: int | None = None,
        width: int | None = None,
        height: int | None = None,
    ) -> None:
        """Makes this window a static window, attached to a Screen.

        Values left unspecified are taken from the existing window state.
        """
        self.defunct = True

    @expose_command()
    def center(self) -> None:
        """Centers a floating window on the screen."""
        if not self.floating:
            return

        if not (self.group and self.group.screen):
            return

        screen = self.group.screen

        x = screen.x + (screen.width - self.width) // 2
        y = screen.y + (screen.height - self.height) // 2

        self.place(
            x,
            y,
            self.width,
            self.height,
            self.borderwidth,
            self.bordercolor,
            above=True,
            respect_hints=True,
        )

    def activate(self) -> bool:
        """Focus and raise this window."""
        if self.group is None:
            return False
        # Windows belonging to a scratchpad need to be toggled properly
        if isinstance(self.group, ScratchPad):
            for dropdown in self.group.dropdowns.values():
                if dropdown.window is self:
                    dropdown.show()
                    return True
            return False
        # Normal window activation
        self.qtile.current_screen.set_group(self.group)
        self.group.focus(self)
        self.bring_to_front()
        return True

    def activate_by_config(self) -> None:
        """Activate the window according to focus_on_window_activation setting."""
        focus_behavior = self.qtile.config.focus_on_window_activation
        if focus_behavior == "focus" or callable(focus_behavior) and focus_behavior(self):
            logger.debug("Focusing window (focus_on_window_activation='focus')")
            self.activate()
        elif focus_behavior == "smart":
            if self.group and self.group.screen == self.qtile.current_screen:
                logger.debug("Focusing window (focus_on_window_activation='smart')")
                self.activate()
            else:
                logger.debug("Setting urgent window (focus_on_window_activation='smart')")
                self.urgent = True
                hook.fire("client_urgent_hint_changed", self)
        elif focus_behavior == "urgent":
            logger.debug("Setting urgent window (focus_on_window_activation='urgent')")
            self.urgent = True
            hook.fire("client_urgent_hint_changed", self)
        elif focus_behavior == "never":
            logger.debug("Ignoring focus request (focus_on_window_activation='never')")
        else:
            logger.debug("Invalid value for focus_on_window_activation: %s", focus_behavior)

    def add_config_inhibitors(self) -> None:
        for rule in self.qtile.config.idle_inhibitors:
            if rule.match is None or rule.match.compare(self):
                self.add_idle_inhibitor(rule.when)

    @expose_command()
    def add_idle_inhibitor(self, inhibitor_type: str = "open") -> None:
        """
        Create an inhibitor rule for this window.

        ``inhibitor_type`` should be one of ``"open"``, ``"focus"``, ``"fullscreen"``
        or ``"visible"``. Default value is ``"open"``.
        """
        self.qtile.core.idle_inhibitor_manager.add_window_inhibitor(self, inhibitor_type)

    @expose_command()
    def remove_idle_inhibitor(self) -> None:
        """Remove inhibitor rule for this window."""
        self.qtile.core.idle_inhibitor_manager.remove_window_inhibitor(self)


class Internal(_Window, metaclass=ABCMeta):
    """An Internal window belonging to Qtile."""

    def __repr__(self):
        return f"Internal(wid={self.wid})"

    @abstractmethod
    def create_drawer(self, width: int, height: int) -> Drawer:
        """Create a Drawer that draws to this window."""

    def process_window_expose(self) -> None:
        """Respond to the window being exposed. Required by X11 backend."""

    def process_button_click(self, x: int, y: int, button: int) -> None:
        """Handle a pointer button click."""

    def process_button_release(self, x: int, y: int, button: int) -> None:
        """Handle a pointer button release."""

    def process_pointer_enter(self, x: int, y: int) -> None:
        """Handle the pointer entering the window."""

    def process_pointer_leave(self, x: int, y: int) -> None:
        """Handle the pointer leaving the window."""

    def process_pointer_motion(self, x: int, y: int) -> None:
        """Handle pointer motion within the window."""

    def process_key_press(self, keycode: int) -> None:
        """Handle a key press."""

    def paint_borders(self, color: ColorsType, width: int) -> None:
        """Paint the window borders with the given color(s) and width"""

    def focus(self, warp: bool = True) -> None:
        """Focus this window and optional warp the pointer to it."""


class Static(_Window, metaclass=ABCMeta):
    """A window bound to a screen rather than a group."""

    screen: config.Screen
    x: Any
    y: Any
    width: Any
    height: Any

    def __repr__(self):
        return f"{self.__class__.__name__!s}(name={self.name!r}, wid={self.wid:d})"

    @expose_command()
    def info(self) -> dict:
        """Return a dictionary of info."""
        return dict(
            name=self.name,
            wm_class=self.get_wm_class(),
            x=self.x,
            y=self.y,
            width=self.width,
            height=self.height,
            id=self.wid,
        )


WindowType = Window | Internal | Static



from __future__ import annotations

import abc
import glob
import os
import pickle
import string
from collections import deque

from libqtile import hook, pangocffi, utils
from libqtile.backend.base.window import Window
from libqtile.command.base import CommandObject, SelectError, expose_command
from libqtile.command.client import InteractiveCommandClient
from libqtile.command.interface import CommandError, QtileCommandInterface
from libqtile.core.manager import Qtile
from libqtile.log_utils import logger
from libqtile.widget import base


class AbstractCompleter(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def __init__(self, qtile: CommandObject) -> None:
        pass

    @abc.abstractmethod
    def actual(self) -> str | None:
        pass

    @abc.abstractmethod
    def reset(self) -> None:
        pass

    @abc.abstractmethod
    def complete(self, txt: str, aliases: dict[str, str] | None = None) -> str:
        """
        Perform the requested completion on the given text.

        The completer can optionally support aliases, which map strings to commands. The
        completer should include the aliases in the completion results.
        """
        # pragma: no cover


class NullCompleter(AbstractCompleter):
    def __init__(self, qtile) -> None:
        self.qtile = qtile

    def actual(self) -> str:
        return ""

    def reset(self) -> None:
        pass

    def complete(self, txt: str, _aliases: dict[str, str] | None = None) -> str:
        return txt


class FileCompleter(AbstractCompleter):
    def __init__(self, qtile: Qtile, _testing=False) -> None:
        self._testing = _testing
        self.qtile = qtile
        self.thisfinal = None  # type: str | None
        self.lookup = None  # type: list[tuple[str, str]] | None
        self.reset()

    def actual(self) -> str | None:
        return self.thisfinal

    def reset(self) -> None:
        self.lookup = None

    def complete(self, txt: str, _aliases: dict[str, str] | None = None) -> str:
        """Returns the next completion for txt, or None if there is no completion"""
        if self.lookup is None:
            self.lookup = []
            if txt == "" or txt[0] not in "~/":
                txt = "~/" + txt
            path = os.path.expanduser(txt)
            if os.path.isdir(path):
                files = glob.glob(os.path.join(path, "*"))
                prefix = txt
            else:
                files = glob.glob(path + "*")
                prefix = os.path.dirname(txt)
                prefix = prefix.rstrip("/") or "/"
            for f in files:
                display = os.path.join(prefix, os.path.basename(f))
                if os.path.isdir(f):
                    display += "/"
                self.lookup.append((display, f))
                self.lookup.sort()
            self.offset = -1
            self.lookup.append((txt, txt))
        self.offset += 1
        if self.offset >= len(self.lookup):
            self.offset = 0
        ret = self.lookup[self.offset]
        self.thisfinal = ret[1]
        return ret[0]


class QshCompleter(AbstractCompleter):
    def __init__(self, qtile: Qtile) -> None:
        q = QtileCommandInterface(qtile)
        self.client = InteractiveCommandClient(q)
        self.thisfinal = None  # type: str | None
        self.reset()

    def actual(self) -> str | None:
        return self.thisfinal

    def reset(self) -> None:
        self.lookup = None  # type: list[tuple[str, str]] | None
        self.path = ""
        self.offset = -1

    def complete(self, txt: str, _aliases: dict[str, str] | None = None) -> str:
        txt = txt.lower()
        if self.lookup is None:
            self.lookup = []
            path = txt.split(".")[:-1]
            self.path = ".".join(path)
            term = txt.split(".")[-1]
            if len(self.path) > 0:
                self.path += "."

            contains_cmd = f"self.client.{self.path}_contains"
            try:
                contains = eval(contains_cmd)
            except AttributeError:
                contains = []
            for obj in contains:
                if obj.lower().startswith(term):
                    self.lookup.append((obj, obj))

            commands_cmd = f"self.client.{self.path}commands()"
            try:
                commands = eval(commands_cmd)
            except (CommandError, AttributeError):
                commands = []
            for cmd in commands:
                if cmd.lower().startswith(term):
                    self.lookup.append((cmd + "()", cmd + "()"))

            self.offset = -1
            self.lookup.append((term, term))

        self.offset += 1
        if self.offset >= len(self.lookup):
            self.offset = 0
        ret = self.lookup[self.offset]
        self.thisfinal = self.path + ret[0]
        return self.path + ret[0]


class GroupCompleter(AbstractCompleter):
    def __init__(self, qtile: Qtile) -> None:
        self.qtile = qtile
        self.thisfinal = None  # type: str | None
        self.lookup = None  # type: list[tuple[str, str]] | None
        self.offset = -1

    def actual(self) -> str | None:
        """Returns the current actual value"""
        return self.thisfinal

    def reset(self) -> None:
        self.lookup = None
        self.offset = -1

    def complete(self, txt: str, _aliases: dict[str, str] | None = None) -> str:
        """Returns the next completion for txt, or None if there is no completion"""
        txt = txt.lower()
        if not self.lookup:
            self.lookup = []
            for group in self.qtile.groups_map.keys():
                if group.lower().startswith(txt):
                    self.lookup.append((group, group))

            self.lookup.sort()
            self.offset = -1
            self.lookup.append((txt, txt))

        self.offset += 1
        if self.offset >= len(self.lookup):
            self.offset = 0
        ret = self.lookup[self.offset]
        self.thisfinal = ret[1]
        return ret[0]


class WindowCompleter(AbstractCompleter):
    def __init__(self, qtile: Qtile) -> None:
        self.qtile = qtile
        self.thisfinal = None  # type: str | None
        self.lookup = None  # type: list[tuple[str, str]] | None
        self.offset = -1

    def actual(self) -> str | None:
        """Returns the current actual value"""
        return self.thisfinal

    def reset(self) -> None:
        self.lookup = None
        self.offset = -1

    def complete(self, txt: str, _aliases: dict[str, str] | None = None) -> str:
        """Returns the next completion for txt, or None if there is no completion"""
        if self.lookup is None:
            self.lookup = []
            for wid, window in self.qtile.windows_map.items():
                if not isinstance(window, Window):
                    continue
                if window.group and window.name.lower().startswith(txt):
                    self.lookup.append((window.name, str(wid)))

            self.lookup.sort()
            self.offset = -1
            self.lookup.append((txt, txt))

        self.offset += 1
        if self.offset >= len(self.lookup):
            self.offset = 0
        ret = self.lookup[self.offset]
        self.thisfinal = ret[1]
        return ret[0]


class CommandCompleter:
    """
    Parameters
    ==========
    _testing :
        disables reloading of the lookup table to make testing possible.
    """

    DEFAULTPATH = "/bin:/usr/bin:/usr/local/bin"

    def __init__(self, qtile: Qtile, _testing=False):
        self.lookup = None  # type: list[tuple[str, str]] | None
        self.offset = -1
        self.thisfinal = None  # type: str | None
        self._testing = _testing

    def actual(self) -> str | None:
        """Returns the current actual value"""
        return self.thisfinal

    def executable(self, fpath: str):
        return os.access(fpath, os.X_OK)

    def reset(self) -> None:
        self.lookup = None
        self.offset = -1

    def complete(self, txt: str, aliases: dict[str, str] | None = None) -> str:
        """Returns the next completion for txt, or None if there is no completion"""
        if self.lookup is None:
            # Lookup is a set of (display value, actual value) tuples.
            self.lookup = []
            if txt and txt[0] in "~/":
                path = os.path.expanduser(txt)
                if os.path.isdir(path):
                    files = glob.glob(os.path.join(path, "*"))
                    prefix = txt
                else:
                    files = glob.glob(path + "*")
                    prefix = os.path.dirname(txt)
                prefix = prefix.rstrip("/") or "/"
                for f in files:
                    if self.executable(f):
                        display = os.path.join(prefix, os.path.basename(f))
                        if os.path.isdir(f):
                            display += "/"
                        self.lookup.append((display, f))
            else:
                dirs = os.environ.get("PATH", self.DEFAULTPATH).split(":")
                for d in dirs:
                    try:
                        d = os.path.expanduser(d)
                        for cmd in glob.iglob(os.path.join(d, f"{txt}*")):
                            if self.executable(cmd):
                                self.lookup.append(
                                    (os.path.basename(cmd), cmd),
                                )
                    except OSError:
                        pass

            if aliases:
                for alias in aliases:
                    if alias.startswith(txt):
                        self.lookup.append((alias, aliases[alias]))

            self.lookup.sort()
            self.offset = -1
            self.lookup.append((txt, txt))
        self.offset += 1
        if self.offset >= len(self.lookup):
            self.offset = 0
        ret = self.lookup[self.offset]
        self.thisfinal = ret[1]
        return ret[0]


class Prompt(base._TextBox):
    """A widget that prompts for user input

    Input should be started using the ``.start_input()`` method on this class.
    """

    completers = {
        "file": FileCompleter,
        "qshell": QshCompleter,
        "cmd": CommandCompleter,
        "group": GroupCompleter,
        "window": WindowCompleter,
        None: NullCompleter,
    }

    defaults = [
        ("cursor", True, "Show a cursor"),
        ("cursorblink", 0.5, "Cursor blink rate. 0 to disable."),
        ("cursor_color", "bef098", "Color for the cursor and text over it."),
        ("prompt", "{prompt}: ", "Text displayed at the prompt"),
        ("record_history", True, "Keep a record of executed commands"),
        ("max_history", 100, "Commands to keep in history. 0 for no limit."),
        ("ignore_dups_history", False, "Don't store duplicates in history"),
        (
            "bell_style",
            "audible",
            "Alert at the begin/end of the command history. "
            + "Possible values: 'audible' (X11 only), 'visual' and None.",
        ),
        ("visual_bell_color", "ff0000", "Color for the visual bell (changes prompt background)."),
        ("visual_bell_time", 0.2, "Visual bell duration (in seconds)."),
    ]

    def __init__(self, **config) -> None:
        base._TextBox.__init__(self, "", **config)
        self.add_defaults(Prompt.defaults)
        self.active = False
        self.completer = None  # type: AbstractCompleter | None
        self.aliases: dict[str, str] | None = None

        # If history record is on, get saved history or create history record
        if self.record_history:
            self.history_path = os.path.join(utils.get_cache_dir(), "prompt_history")
            if os.path.exists(self.history_path):
                with open(self.history_path, "rb") as f:
                    try:
                        self.history = pickle.load(f)
                        if self.ignore_dups_history:
                            self._dedup_history()
                    except:  # noqa: E722
                        # unfortunately, pickle doesn't wrap its errors, so we
                        # can't detect what's a pickle error and what's not.
                        logger.exception("failed to load prompt history")
                        self.history = {
                            x: deque(maxlen=self.max_history) for x in self.completers
                        }

                    # self.history of size does not match.
                    if len(self.history) != len(self.completers):
                        self.history = {
                            x: deque(maxlen=self.max_history) for x in self.completers
                        }

                    if self.max_history != self.history[list(self.history)[0]].maxlen:
                        self.history = {
                            x: deque(self.history[x], self.max_history) for x in self.completers
                        }
            else:
                self.history = {x: deque(maxlen=self.max_history) for x in self.completers}

    def _configure(self, qtile, bar) -> None:
        self.markup = True
        base._TextBox._configure(self, qtile, bar)

        def f(win):
            if self.active and not win == self.bar.window:
                self._unfocus()

        def prevent_focus_steal(group, win):
            if self.active:
                win.can_steal_focus = False

        hook.subscribe.group_window_add(prevent_focus_steal)
        hook.subscribe.client_focus(f)

        # Define key handlers (action to do when a specific key is hit)
        keyhandlers = {
            "Tab": self._trigger_complete,
            "BackSpace": self._delete_char(),
            "Delete": self._delete_char(False),
            "KP_Delete": self._delete_char(False),
            "Escape": self._unfocus,
            "Return": self._send_cmd,
            "KP_Enter": self._send_cmd,
            "Up": self._get_prev_cmd,
            "KP_Up": self._get_prev_cmd,
            "Down": self._get_next_cmd,
            "KP_Down": self._get_next_cmd,
            "Left": self._move_cursor(),
            "KP_Left": self._move_cursor(),
            "Right": self._move_cursor("right"),
            "KP_Right": self._move_cursor("right"),
        }
        self.keyhandlers = {qtile.core.keysym_from_name(k): v for k, v in keyhandlers.items()}
        printables = {x: self._write_char for x in range(127) if chr(x) in string.printable}
        self.keyhandlers.update(printables)
        self.tab = qtile.core.keysym_from_name("Tab")

        self.bell_style: str
        if self.bell_style == "audible" and qtile.core.name != "x11":
            self.bell_style = "visual"
            logger.warning("Prompt widget only supports audible bell under X11")
        if self.bell_style == "visual":
            self.original_background = self.background

    def start_input(
        self,
        prompt,
        callback,
        complete=None,
        strict_completer=False,
        allow_empty_input=False,
        aliases: dict[str, str] | None = None,
    ) -> None:
        """Run the prompt

        Displays a prompt and starts to take one line of keyboard input from
        the user. When done, calls the callback with the input string as
        argument. If history record is enabled, also allows to browse between
        previous commands with ↑ and ↓, and execute them (untouched or
        modified). When history is exhausted, fires an alert. It tries to
        mimic, in some way, the shell behavior.

        Parameters
        ==========
        complete :
            Tab-completion. Can be None, "cmd", "file", "group", "qshell" or
            "window".
        prompt :
            text displayed at the prompt, e.g. "spawn: "
        callback :
            function to call with returned value.
        complete :
            completer to use.
        strict_completer :
            When True the return value wil be the exact completer result where
            available.
        allow_empty_input :
            When True, an empty value will still call the callback function
        aliases :
            Dictionary mapping aliases to commands. If the entered command is a key in
            this dict, the command it maps to will be executed instead.
        """

        if self.cursor and self.cursorblink and not self.active:
            self.timeout_add(self.cursorblink, self._blink)
        self.display = self.prompt.format(prompt=prompt)
        self.display = pangocffi.markup_escape_text(self.display)
        self.active = True
        self.user_input = ""
        self.archived_input = ""
        self.show_cursor = self.cursor
        self.cursor_position = 0
        self.callback = callback
        self.aliases = aliases
        self.completer = self.completers[complete](self.qtile)
        self.strict_completer = strict_completer
        self.allow_empty_input = allow_empty_input
        self._update()
        self.bar.widget_grab_keyboard(self)
        if self.record_history:
            self.completer_history = self.history[complete]
            self.position = len(self.completer_history)

    def _blink(self) -> None:
        self.show_cursor = not self.show_cursor
        self._update()
        if self.active:
            self.timeout_add(self.cursorblink, self._blink)

    def _highlight_text(self, text) -> str:
        color = utils.hex(self.cursor_color)
        text = f'<span foreground="{color}">{text}</span>'
        if self.show_cursor:
            text = f"<u>{text}</u>"
        return text

    def _update(self) -> None:
        if self.active:
            self.text = self.archived_input or self.user_input
            cursor = pangocffi.markup_escape_text(" ")
            if self.cursor_position < len(self.text):
                txt1 = self.text[: self.cursor_position]
                txt2 = self.text[self.cursor_position]
                txt3 = self.text[self.cursor_position + 1 :]
                for text in (txt1, txt2, txt3):
                    text = pangocffi.markup_escape_text(text)
                txt2 = self._highlight_text(txt2)
                self.text = f"{txt1}{txt2}{txt3}{cursor}"
            else:
                self.text = pangocffi.markup_escape_text(self.text)
                self.text += self._highlight_text(cursor)
            self.text = self.display + self.text
        else:
            self.text = ""
        self.bar.draw()

    def _trigger_complete(self) -> None:
        # Trigger the auto completion in user input
        assert self.completer is not None
        self.user_input = self.completer.complete(self.user_input, self.aliases)
        self.cursor_position = len(self.user_input)

    def _history_to_input(self) -> None:
        # Move actual command (when exploring history) to user input and update
        # history position (right after the end)
        if self.archived_input:
            self.user_input = self.archived_input
            self.archived_input = ""
            self.position = len(self.completer_history)

    def _insert_before_cursor(self, charcode) -> None:
        # Insert a character (given their charcode) in input, before the cursor
        txt1 = self.user_input[: self.cursor_position]
        txt2 = self.user_input[self.cursor_position :]
        self.user_input = txt1 + chr(charcode) + txt2
        self.cursor_position += 1

    def _delete_char(self, backspace=True):
        # Return a function that deletes character from the input text.
        # If backspace is True, function will emulate backspace, else Delete.
        def f():
            self._history_to_input()
            step = -1 if backspace else 0
            if not backspace and self.cursor_position == len(self.user_input):
                self._alert()
            elif len(self.user_input) > 0 and self.cursor_position + step > -1:
                txt1 = self.user_input[: self.cursor_position + step]
                txt2 = self.user_input[self.cursor_position + step + 1 :]
                self.user_input = txt1 + txt2
                if step:
                    self.cursor_position += step
            else:
                self._alert()

        return f

    def _write_char(self):
        # Add pressed (legal) char key to user input.
        # No LookupString in XCB... oh, the shame! Unicode users beware!
        self._history_to_input()
        self._insert_before_cursor(self.key)

    def _unfocus(self):
        # Remove focus from the widget
        self.active = False
        self._update()
        self.bar.widget_ungrab_keyboard()

    def _send_cmd(self):
        # Send the prompted text for execution
        self._unfocus()
        if self.strict_completer:
            self.user_input = self.actual_value or self.user_input
            del self.actual_value
        self._history_to_input()
        if self.user_input or self.allow_empty_input:
            # If history record is activated, also save command in history
            if self.record_history:
                # ensure no dups in history
                if self.ignore_dups_history and (self.user_input in self.completer_history):
                    self.completer_history.remove(self.user_input)
                    self.position -= 1

                self.completer_history.append(self.user_input)

                if self.position < self.max_history:
                    self.position += 1
                os.makedirs(os.path.dirname(self.history_path), exist_ok=True)
                with open(self.history_path, mode="wb") as f:
                    pickle.dump(self.history, f, protocol=2)
            self.callback(self.user_input)

    def _alert(self):
        # Fire an alert (audible or visual), if bell style is not None.
        if self.bell_style == "audible":
            self.qtile.core.conn.conn.core.Bell(0)
        elif self.bell_style == "visual":
            self.background = self.visual_bell_color
            self.timeout_add(self.visual_bell_time, self._stop_visual_alert)

    def _stop_visual_alert(self):
        self.background = self.original_background
        self._update()

    def _get_prev_cmd(self):
        # Get the previous command in history.
        # If there isn't more previous commands, ring system bell
        if self.record_history:
            if not self.position:
                self._alert()
            else:
                self.position -= 1
                self.archived_input = self.completer_history[self.position]
                self.cursor_position = len(self.archived_input)

    def _get_next_cmd(self):
        # Get the next command in history.
        # If the last command was already reached, ring system bell.
        if self.record_history:
            if self.position == len(self.completer_history):
                self._alert()
            elif self.position < len(self.completer_history):
                self.position += 1
                if self.position == len(self.completer_history):
                    self.archived_input = ""
                else:
                    self.archived_input = self.completer_history[self.position]
                self.cursor_position = len(self.archived_input)

    def _cursor_to_left(self):
        # Move cursor to left, if possible
        if self.cursor_position:
            self.cursor_position -= 1
        else:
            self._alert()

    def _cursor_to_right(self):
        # move cursor to right, if possible
        command = self.archived_input or self.user_input
        if self.cursor_position < len(command):
            self.cursor_position += 1
        else:
            self._alert()

    def _move_cursor(self, direction="left"):
        # Move the cursor to left or right, according to direction
        if direction == "left":
            return self._cursor_to_left
        elif direction == "right":
            return self._cursor_to_right

    def _get_keyhandler(self, k):
        # Return the action (a function) to do according the pressed key (k).
        self.key = k
        if k in self.keyhandlers:
            if k != self.tab:
                self.actual_value = self.completer.actual()
                self.completer.reset()
            return self.keyhandlers[k]

    def process_key_press(self, keysym: int):
        """Key press handler for the minibuffer.

        Currently only supports ASCII characters.
        """
        handle_key = self._get_keyhandler(keysym)

        if handle_key:
            handle_key()
            del self.key
        self._update()

    @expose_command()
    def fake_keypress(self, key: str) -> None:
        self.process_key_press(self.qtile.core.keysym_from_name(key))

    @expose_command()
    def info(self):
        """Returns a dictionary of info for this object"""
        return dict(
            name=self.name,
            width=self.width,
            text=self.text,
            active=self.active,
        )

    @expose_command()
    def exec_general(self, prompt, object_name, cmd_name, selector=None, completer=None):
        """
        Execute a cmd of any object. For example layout, group, window, widget
        , etc with a string that is obtained from start_input.

        Parameters
        ==========
        prompt :
            Text displayed at the prompt.
        object_name :
            Name of a object in Qtile. This string has to be 'layout', 'widget',
            'bar', 'window' or 'screen'.
        cmd_name :
            Execution command of selected object using object_name and selector.
        selector :
            This value select a specific object within a object list that is
            obtained by object_name.
            If this value is None, current object is selected. e.g. current layout,
            current window and current screen.
        completer:
            Completer to use.

        config example:
            Key([alt, 'shift'], 'a',
                lazy.widget['prompt'].exec_general(
                    'section(add)',
                    'layout',
                    'add_section'))
        """
        try:
            obj = self.qtile.select([(object_name, selector)])
        except SelectError:
            logger.warning("cannot select a object")
            return
        cmd = obj.command(cmd_name)
        if not cmd:
            logger.warning("command not found")
            return

        def f(args):
            if args:
                cmd(args)

        self.start_input(prompt, f, completer)

    def _dedup_history(self):
        """Filter the history deque, clearing all duplicate values."""
        self.history = {x: self._dedup_deque(self.history[x]) for x in self.completers}

    def _dedup_deque(self, dq):
        return deque(_LastUpdatedOrdereddict.fromkeys(dq))


class _LastUpdatedOrdereddict(dict):
    """Store items in the order the keys were last added."""

    def __setitem__(self, key, value):
        if key in self:
            del self[key]
        super().__setitem__(key, value)



from __future__ import annotations

from math import log

import psutil

from libqtile.widget import base


class Net(base.InLoopPollText):
    """
    Displays interface down and up speed

    The following fields are available in the `format` string:

    - ``interface``: name of the interface
    - ``down``: download speed
    - ``down_suffix``: suffix for the download speed
    - ``down_cumulative``: cumulative download traffic
    - ``down_cumulative_suffix``: suffix for the cumulative download traffic
    - ``up``: upload speed
    - ``up_suffix``: suffix for the upload speed
    - ``up_cumulative``: cumulative upload traffic
    - ``up_cumulative_suffix``: suffix for the cumulative upload traffic
    - ``total``: total speed
    - ``total_suffix``: suffix for the total speed
    - ``total_cumulative``: cumulative total traffic
    - ``total_cumulative_suffix``: suffix for the cumulative total traffic

    Widget requirements: psutil_.

    .. _psutil: https://pypi.org/project/psutil/
    """

    defaults = [
        (
            "format",
            "{interface}: {down:6.2f}{down_suffix:<2}\u2193\u2191{up:6.2f}{up_suffix:<2}",
            "Display format of down/upload/total speed of given interfaces",
        ),
        (
            "missing_interface",
            "{interface} not found",
            "Display when interface is missing/unavailable.",
        ),
        (
            "interface",
            None,
            "List of interfaces or single NIC as string to monitor, \
            None to display all active NICs combined",
        ),
        ("update_interval", 1, "The update interval."),
        ("use_bits", False, "Use bits instead of bytes per second?"),
        ("prefix", None, "Use a specific prefix for the unit of the speed."),
        (
            "cumulative_prefix",
            None,
            "Use a specific prefix for the unit of the cumulative traffic.",
        ),
    ]

    def __init__(self, **config):
        base.InLoopPollText.__init__(self, "", **config)
        self.add_defaults(Net.defaults)

        self.factor = 1000.0
        self.allowed_prefixes = ["", "k", "M", "G", "T", "P", "E", "Z", "Y"]

        if self.use_bits:
            self.base_unit = "b"
            self.byte_multiplier = 8
        else:
            self.base_unit = "B"
            self.byte_multiplier = 1

        self.units = list(map(lambda p: p + self.base_unit, self.allowed_prefixes))

        if not isinstance(self.interface, list):
            if self.interface is None:
                self.interface = ["all"]
            elif isinstance(self.interface, str):
                self.interface = [self.interface]
            else:
                raise AttributeError(
                    f"Invalid Argument passed: {self.interface}\nAllowed Types: list, str, None"
                )
        self.stats = self.get_stats()

    def convert_b(self, num_bytes: float, prefix: str | None = None) -> tuple[float, str]:
        """Converts the number of bytes to the correct unit"""

        num_bytes *= self.byte_multiplier

        if prefix is None:
            if num_bytes > 0:
                power = int(log(num_bytes) / log(self.factor))
                power = min(power, len(self.units) - 1)
            else:
                power = 0
        else:
            power = self.allowed_prefixes.index(prefix)

        converted_bytes = num_bytes / self.factor**power
        unit = self.units[power]

        return converted_bytes, unit

    def get_stats(self):
        interfaces = {}
        if self.interface == ["all"]:
            net = psutil.net_io_counters(pernic=False)
            interfaces["all"] = {
                "down": net.bytes_recv,
                "up": net.bytes_sent,
                "total": net.bytes_recv + net.bytes_sent,
            }
            return interfaces
        else:
            net = psutil.net_io_counters(pernic=True)
            for iface in net:
                down = net[iface].bytes_recv
                up = net[iface].bytes_sent
                interfaces[iface] = {
                    "down": down,
                    "up": up,
                    "total": down + up,
                }
            return interfaces

    def poll(self):
        ret_stat = []
        new_stats = self.get_stats()
        for intf in self.interface:
            if intf not in new_stats:
                ret_stat.append(self.missing_interface.format(interface=intf))
                continue

            # This is for any interfaces that appear after the widget has started.
            # In that scenario, new_stats[intf] will exist but self.stats[intf] would result
            # in a KeyError, so we populate it now.
            if intf not in self.stats:
                self.stats[intf] = new_stats[intf]

            down = new_stats[intf]["down"] - self.stats[intf]["down"]
            up = new_stats[intf]["up"] - self.stats[intf]["up"]
            total = new_stats[intf]["total"] - self.stats[intf]["total"]

            down = down / self.update_interval
            up = up / self.update_interval
            total = total / self.update_interval
            down, down_suffix = self.convert_b(down, self.prefix)
            down_cumulative, down_cumulative_suffix = self.convert_b(
                new_stats[intf]["down"], self.cumulative_prefix
            )
            up, up_suffix = self.convert_b(up, self.prefix)
            up_cumulative, up_cumulative_suffix = self.convert_b(
                new_stats[intf]["up"], self.cumulative_prefix
            )
            total, total_suffix = self.convert_b(total, self.prefix)
            total_cumulative, total_cumulative_suffix = self.convert_b(
                new_stats[intf]["total"], self.cumulative_prefix
            )
            self.stats[intf] = new_stats[intf]
            ret_stat.append(
                self.format.format(
                    interface=intf,
                    down=down,
                    down_suffix=down_suffix,
                    down_cumulative=down_cumulative,
                    down_cumulative_suffix=down_cumulative_suffix,
                    up=up,
                    up_suffix=up_suffix,
                    up_cumulative=up_cumulative,
                    up_cumulative_suffix=up_cumulative_suffix,
                    total=total,
                    total_suffix=total_suffix,
                    total_cumulative=total_cumulative,
                    total_cumulative_suffix=total_cumulative_suffix,
                )
            )

        return " ".join(ret_stat)
